"""Databases.py: This file contains MINE database classes including database
loading functions."""
import ast
import datetime
import os
import sys
import platform
from shutil import move
from subprocess import call

import pymongo
from pymongo.errors import ServerSelectionTimeoutError
from rdkit.Chem import AllChem

from minedatabase import utils
from minedatabase.NP_Score import npscorer as np


def establish_db_client():
    """This establishes a mongo database client in various environments"""
    try:
        # Special case for working on a sapphire node
        if 'node' in platform.node():
            client = pymongo.MongoClient(host='master',
                                         serverSelectionTimeoutMS=5)
        # Special case for working on a SEED cluster
        elif 'bio' in platform.node() or platform.node() == 'twig' \
                or platform.node() == 'branch':
            client = pymongo.MongoClient(host='branch',
                                         serverSelectionTimeoutMS=5)
            admin = client['admin']
            admin.authenticate('worker', 'bnice14bot')
        # Local database
        else:
            client = pymongo.MongoClient(serverSelectionTimeoutMS=5)
    except ServerSelectionTimeoutError:
        raise IOError("Failed to load database client. Please verify that "
                      "mongod is running")
    return client


class MINE:
    """
    This class basically exposes the underlying mongo database to manipulation
    but also defines expected database structure.
    """
    def __init__(self, name):
        client = establish_db_client()
        db = client[name]
        self._db = db
        self.name = name
        self.meta_data = db.meta_data
        self.compounds = db.compounds
        self.reactions = db.reactions
        self.operators = db.operators
        self.models = db.models
        self.np_model = None
        self.id_db = client['UniversalMINE']
        self._mass_cache = {}  # for rapid calculation of reaction mass change

    def add_rxn_pointers(self):
        """Add links to the reactions that each compound participates in
        allowing for users to follow paths in the network"""
        reactions_count = self.reactions.count()
        print("Linking compounds to %s reactions" % reactions_count)
        for reaction in self.reactions.find().batch_size(500):
            # Update pointers for compounds that are reactants
            for compound in reaction['Reactants']:
                self.compounds.update({"_id": compound["c_id"]},
                                      {'$push': {
                                          "Reactant_in": reaction['_id']}})
            # Update pointers for compounds that are products
            for compound in reaction['Products']:
                self.compounds.update({"_id": compound["c_id"]},
                                      {'$push': {
                                          "Product_of": reaction['_id']}})
        # Write to log file
        self.meta_data.insert({"Timestamp": datetime.datetime.now(), "Action":
                               "Add Reaction Pointers"})

    def add_compound_sources(self, rxn_key_type="_id"):
        """Adds a field detailing the compounds and reactions from which a
        compound is produced. This enables filtering search results to only
        include compounds which are relevant to a specified metabolic context.
        This is different from add_rxn_pointers in that this provides the
        precursor compounds rather than the reactions that the compound
        participates in."""
        for compound in self.compounds.find({"Sources": {"$exists": 0}}):
            compound['Sources'] = []

            for reaction in self.reactions.find(
                    {"Products.c_id": compound[rxn_key_type]}):
                compound['Sources'].append(
                    {"Compounds": [x['c_id'] for x in reaction['Reactants']],
                     "Operators": reaction["Operators"]})
            # If there are sources, then save them and make sure there aren't
            #  too many for a single compound.
            if compound['Sources']:
                try:
                    self.compounds.save(compound)
                except pymongo.errors.DocumentTooLarge:
                    print("Too Many Sources for %s" % compound['SMILES'])
        # Allow for quick searching within database
        self.compounds.ensure_index([("Sources.Compound", pymongo.ASCENDING),
                                     ("Sources.Operators", pymongo.ASCENDING)])
        # Write to log file
        self.meta_data.insert({"Timestamp": datetime.datetime.now(), "Action":
                               "Add Compound Source field"})

    def add_reaction_mass_change(self, reaction=None):
        """Calculate the change in mass between reactant and product compounds.
            This is useful for discovering compounds in molecular networking
        :param reaction: If specified, the function returns the mass change for
            the supplied reaction. If false, it calculates the mass change for
            all reactions missing this value in the database.
        :type reaction: str
        :return: If reaction specified, the mass change
        :rtype: float
        """
        def _get_mass(_id):
            if _id not in self._mass_cache:
                try:
                    self._mass_cache['_id'] = self.compounds.find_one(
                        {"_id": _id}, {'Mass': 1})['Mass']
                except (TypeError, KeyError):
                    raise ValueError('A mass value for %s was not found in '
                                     'compounds collection' % _id)
            return self._mass_cache['_id']

        def _get_mass_change(rxn):
            if isinstance(rxn['Reactants'][0], dict):
                key = 'c_id'
            else:
                key = 1
            start_mass = _get_mass(rxn['Reactants'][0][key])
            return [_get_mass(product[key]) - start_mass for product
                    in rxn['Products'] if product[key][0] == 'C']

        if reaction:
            return _get_mass_change(reaction)

        bulk = self.reactions.initialize_unordered_bulk_op()
        reactions = self.reactions.find({"Mass_Change": {'$exists': 0}})
        for rxn in reactions:
            try:
                bulk.find({'_id': rxn['_id']}).update({'$set': {
                    'Mass_Change': _get_mass_change(rxn)}})
            except ValueError as e:
                print(e.args[0])
                print('Failed to parse rxn %s' % rxn['_id'])
        bulk.execute()

    def generate_image_files(self, path, query=None, dir_depth=0,
                             img_type='svg:-a,nosource,w500,h500',
                             convert_r=False):
        """Generates image files for compounds in database using ChemAxon's
        MolConvert.

        :param path: Target directory for image files
        :type path: str
        :param query: A query to limit the number of files generated
        :type query: dict
        :param dir_depth: The number of directory levels to split the
            compounds into for files system efficiency. Ranges from 0 (all in
            top level directory to the length of the file name (40 for MINE
            hashes)
        :type dir_depth: int
        :param img_type: The type of image file to be generated. See molconvert
            documentation for valid options
        :type img_type: str

        """
        ids = []
        extension = img_type.split(":")[0]
        structure_file = os.path.join(path, 'tmp.smiles')
        if not query:
            query = {}

        if not os.path.exists(path):
            if sys.platform == "linux":
                os.system(f"sudo mkdir {path} -m 777")
            else:
                os.mkdir(path)
        with open(structure_file, 'w') as outfile:
            for comp in self.compounds.find(query, {'SMILES': 1}):
                if convert_r:
                    outfile.write("%s\n" % comp['SMILES'].replace('R', "*"))
                else:
                    outfile.write("%s\n" % comp['SMILES'])
                ids.append(comp['_id'])
        if platform.system() == 'Windows':
            rc = call("molconvert -mo \"%s/.%s\" %s \"%s\""
                      % (path, extension, img_type, structure_file),
                      shell=True)
        else:
            rc = call(["molconvert -mo %s/.%s %s %s"
                       % (path, extension, img_type, structure_file)],
                      shell=True)
        if rc:
            raise RuntimeError("molconvert returned %s" % rc)
        os.remove(structure_file)

        for i, _id in enumerate(ids):
            old = os.path.join(path, "%s.%s" % ((i + 1), extension))
            new = path
            for j in range(0, dir_depth):
                new = os.path.join(new, _id[j])
            if not os.path.exists(new):
                os.makedirs(new)
            new = os.path.join(new, _id + '.' + extension)
            if os.path.isfile(old):
                move(old, new)

    def build_indexes(self):
        """Builds indexes for efficient querying of MINE databases"""
        self.compounds.create_index([('Mass', pymongo.ASCENDING),
                                     ('Charge', pymongo.ASCENDING),
                                     ('DB_links.Model_SEED',
                                      pymongo.ASCENDING)])
        self.compounds.create_index([('Names', 'text'), ('Pathways', 'text')])
        self.compounds.create_index("DB_links.Model_SEED")
        self.compounds.create_index("DB_links.KEGG")
        self.compounds.create_index("MACCS")
        self.compounds.create_index("len_MACCS")
        self.compounds.create_index("RDKit")
        self.compounds.create_index("len_RDKit")
        self.compounds.create_index("Inchikey")
        self.compounds.create_index("MINE_id")
        self.compounds.create_index("Names")
        self.reactions.create_index("Reactants.c_id")
        self.reactions.create_index("Products.c_id")
        self.meta_data.insert_one({"Timestamp": datetime.datetime.now(),
                                   "Action": "Database indexes built"})

    def link_to_external_database(self, external_database, compound=None,
                                  match_field="Inchikey", fields_to_copy=None):
        """This function looks for matching compounds in other databases (i.e.
        PubChem) and adds links where found.

        :param external_database: The name of the database to search for
            matching compounds
        :type external_database: str
        :param compound: The compound to search for external links. If none,
            link all compounds in the database.
        :type compound: dict
        :param match_field: The field to search on for matching compounds
        :type match_field: str
        :param fields_to_copy: Data to copy into the mine database. The first
            field is the field name in the external database. The second field
            is the field name in the MINE database where the data will be
            copied.
        :type fields_to_copy: list(tuple)
        """
        if compound:
            ext = MINE(external_database)
            projection = dict([("_id", 0,)] + [(x[0], 1,) for x
                                               in fields_to_copy])
            # Find compounds that have same name in another database
            for ext_comp in ext.compounds.find(
                    {match_field: compound[match_field]}, projection):
                for field in fields_to_copy:
                    if field[0] in ext_comp:
                        # dict_merge merges two dictionaries using sets to
                        # avoid duplicate values
                        utils.dict_merge(compound, utils.save_dotted_field(
                            field[1], utils.get_dotted_field(ext_comp,
                                                             field[0])))
            return utils.convert_sets_to_lists(compound)

        # If compound is None, link all compounds in database
        else:
            for comp in self.compounds.find():
                self.compounds.save(self.link_to_external_database(
                    external_database, compound=comp, match_field=match_field,
                    fields_to_copy=fields_to_copy))

    def insert_compound(self, mol_object, compound_dict=None, bulk=None,
                        kegg_db="KEGG", pubchem_db='PubChem-8-28-2015',
                        modelseed_db='ModelSEED'):
        """This class saves a RDKit Molecule as a compound entry in the MINE.
        Calculates necessary fields for API and includes additional
        information passed in the compound dict. Overwrites preexisting
        compounds in MINE on _id collision.

        :param mol_object: The compound to be stored
        :type mol_object: RDKit Mol object
        :param compound_dict: Additional information about the compound to be
            stored. Overwritten by calculated values.
        :type compound_dict: dict
        :param bulk: A pymongo bulk operation object. If None, reaction is
         immediately inserted in the database
        :param kegg_db: The ID of the KEGG Mongo database
        :type kegg_db: str
        :param pubchem_db: The ID of the PubChem Mongo database
        :type pubchem_db: str
        :param modelseed_db: The ID of the ModelSEED Mongo database
        :type modelseed_db: str
        :return: The hashed _id of the compound
        :rtype: str
        """

        if compound_dict is None:
            compound_dict = {}

        # Store all different representations of the molecule (SMILES, Formula,
        #  InChI key, etc.) as well as its properties in a dictionary
        compound_dict['SMILES'] = AllChem.MolToSmiles(mol_object, True)
        compound_dict['Inchi'] = AllChem.MolToInchi(mol_object)
        compound_dict['Inchikey'] = AllChem.InchiToInchiKey(
            compound_dict['Inchi'])
        compound_dict['Mass'] = AllChem.CalcExactMolWt(mol_object)
        compound_dict['Formula'] = AllChem.CalcMolFormula(mol_object)
        compound_dict['Charge'] = AllChem.GetFormalCharge(mol_object)
        # Get indices where bits are 1
        compound_dict['MACCS'] = list(AllChem.GetMACCSKeysFingerprint(
            mol_object).GetOnBits())
        compound_dict['len_MACCS'] = len(compound_dict['MACCS'])
        # Get indices where bits are 1
        compound_dict['RDKit'] = list(AllChem.RDKFingerprint(
            mol_object).GetOnBits())
        compound_dict['len_RDKit'] = len(compound_dict['RDKit'])
        compound_dict['logP'] = AllChem.CalcCrippenDescriptors(mol_object)[0]
        compound_dict['_id'] = utils.compound_hash(
            compound_dict['SMILES'],
            ('Type' in compound_dict
             and compound_dict['Type'] == 'Coreactant'))
        if '_atom_count' in compound_dict:
            del compound_dict['_atom_count']
        # Caching this for rapid reaction mass change calculation
        self._mass_cache[compound_dict['_id']] = compound_dict['Mass']

        # If the compound is a reactant, then make sure the reactant name is
        # in a correct format.
        if "Reactant_in" in compound_dict and isinstance(
                compound_dict['Reactant_in'], str) \
                and compound_dict['Reactant_in']:
            compound_dict['Reactant_in'] = ast.literal_eval(
                compound_dict['Reactant_in'])
        # If the compound is a product, then make sure the reactant name is
        # in a correct format.
        if "Product_of" in compound_dict \
                and isinstance(compound_dict['Product_of'], str) \
                and compound_dict['Product_of']:
            compound_dict['Product_of'] = ast.literal_eval(
                compound_dict['Product_of'])

        # Store links to external databases where compound is present
        if compound_dict['Inchikey']:
            if kegg_db:
                compound_dict = self.link_to_external_database(
                    kegg_db, compound=compound_dict,
                    fields_to_copy=[('Pathways', 'Pathways'),
                                    ('Names', 'Names'),
                                    ('DB_links', 'DB_links'),
                                    ('Enzymes', 'Enzymes')])

            if pubchem_db:
                compound_dict = self.link_to_external_database(
                    pubchem_db, compound=compound_dict,
                    fields_to_copy=[('COMPOUND_CID', 'DB_links.PubChem')])

            if modelseed_db:
                compound_dict = self.link_to_external_database(
                    modelseed_db, compound=compound_dict,
                    fields_to_copy=[('DB_links', 'DB_links')])

        # Calculate natural product likeness score and store in dict
        if not self.np_model:
            self.np_model = np.readNPModel()
        compound_dict["NP_likeness"] = np.scoreMol(mol_object, self.np_model)

        compound_dict = utils.convert_sets_to_lists(compound_dict)
        # Assign an id to the compound
        if self.id_db:
            mine_comp = self.id_db.compounds.find_one(
                {"Inchikey": compound_dict['Inchikey']},
                {'MINE_id': 1, "Pos_CFM_spectra": 1, "Neg_CFM_spectra": 1})
            # If compound already exists in MINE, store its MINE id in the dict
            if mine_comp:
                compound_dict['MINE_id'] = mine_comp['MINE_id']
                if 'Pos_CFM_spectra' in mine_comp:
                    compound_dict['Pos_CFM_spectra'] = \
                        mine_comp['Pos_CFM_spectra']
                if 'Neg_CFM_spectra' in mine_comp:
                    compound_dict['Neg_CFM_spectra'] = \
                        mine_comp['Neg_CFM_spectra']
            # If compound does not exist, create new id based on number of
            # current ids in the MINE
            else:
                compound_dict['MINE_id'] = self.id_db.compounds.count()
                self.id_db.compounds.save(compound_dict)

        # If bulk insertion, upsert (insert and update) the database
        if bulk:
            bulk.find({'_id': compound_dict['_id']}).upsert().\
                replace_one(compound_dict)
        else:
            save_document(self.compounds, compound_dict)
        return compound_dict['_id']

    def insert_reaction(self, reaction_dict, bulk=None):
        """Inserts a reaction into the MINE database and returns _id of the
         reaction in the mine database.

        :param reaction_dict: A dictionary containing "Reactants" and
         "Products" lists of StoichTuples
        :type reaction_dict: dict
        :param bulk: A pymongo bulk operation object. If None, reaction is
         immediately inserted in the database
        :return: The hashed _id of the reaction
        :rtype: str
        """
        reaction_dict['_id'] = utils.rxn2hash(reaction_dict['Reactants'],
                                              reaction_dict['Products'])

        # By converting to a dict, mongo stores the data as objects not
        # arrays allowing for queries by compound hash
        if isinstance(reaction_dict['Reactants'][0], utils.StoichTuple):
            reaction_dict['Reactants'] = [x._asdict() for x
                                          in reaction_dict['Reactants']]
            reaction_dict['Products'] = [x._asdict() for x
                                         in reaction_dict['Products']]

        reaction_dict = utils.convert_sets_to_lists(reaction_dict)
        # If bulk insertion, upsert (insert and update) the database
        if bulk:
            bulk.find({'_id': reaction_dict['_id']}).upsert().\
                replace_one(reaction_dict)
        else:
            save_document(self.reactions, reaction_dict)
        return reaction_dict['_id']

    def map_reactions(self, ext_db, match_field='_id'):
        """Update operators by adding the reactions they participate in.

        :param ext_db: name of MongoDB
        :param match_field: name of reaction field to match based on
        """
        # Update operators by adding the reactions they participate in
        lit_db = MINE(ext_db)
        for lit_rxn in lit_db.reactions.find():
            if match_field == "InChI_hash":
                mine_rxn = self.reactions.find_one(
                    {match_field: {"$regex": lit_rxn[match_field].split('-')[0]
                                   }}, {'Operators': 1})
            else:
                mine_rxn = self.reactions.find_one(
                    {match_field: lit_rxn[match_field]}, {'Operators': 1})
            if mine_rxn:
                for op in mine_rxn['Operators']:
                    self.operators.update(
                        {'_id': op},
                        {"$addToSet": {"Mapped_Rxns": lit_rxn["_id"],
                                       "References": {
                                           "$each": lit_rxn['References']}}})
                lit_db.reactions.update(
                    {"_id": lit_rxn["_id"]},
                    {"$set": {"Mapped_Rules": mine_rxn['Operators']}})


def save_document(collection, doc):
    """Saves document to given MongoDB collection.

    If document _id already exists, replaces it. Else, inserts it.

    Parameters
    ----------
    collection : mongodb.collection
        Collection to save document to.
    doc : dict
        Document to save.
    """
    if '_id' in doc and collection.find_one({'_id': doc['_id']}):
        collection.replace_one({'_id': doc['_id']}, doc)
    else:
        collection.insert_one(doc)



"""Compound_io.py: Functions to load MINE databases from and dump compounds
into common cheminformatics formats"""
import collections
import csv
import datetime
import os
import sys

from rdkit.Chem import AllChem

from minedatabase import utils
from minedatabase.databases import MINE


def export_sdf(mine_db, dir_path, max_compounds=None):
    """Exports compounds from the database as an MDL SDF file

    :param mine_db: a MINE object
    :param dir_path: directory for files
    :param max_compounds: maximum number of compounds per file (defaults to
         unlimited)
    :return:
    """

    # Make sure that all compounds point to all their reactants
    if not mine_db.compounds.find_one({"Product_of": {'$exists': 1}}):
        mine_db.add_rxn_pointers()

    print("Exporting %s compounds from %s as an SDF file" %
          (mine_db.compounds.count(), mine_db.name))
    target = utils.prevent_overwrite(os.path.join(dir_path, mine_db.name)
                                     + "_1.sdf")
    # SDWriter (rdkit) writes Mol objects to SD files
    writer = AllChem.SDWriter(target)
    writer.SetKekulize(True)
    n_files = 1
    for compound in mine_db.compounds.find():
        # Convert SMILES string to Mol object, replacing 'CoA' and 'R' by '*'
        mol = AllChem.MolFromSmiles(compound['SMILES'], True,
                                    {'CoA': '*', 'R': "*"})
        # if Mol object successfully generated, annotate properties
        if mol:
            mol.SetProp('_id', compound['_id'])
            mol.SetProp('Generation', str(compound['Generation']))
            if 'Reactant_in' in compound:
                mol.SetProp('Reactant_in', str(compound['Reactant_in']))
            if 'Product_of' in compound:
                mol.SetProp('Product_of', str(compound['Product_of']))
            writer.write(mol)
            # Start writing a new sdf file if the maximum (set by user) has
            # been reached for the current file
            if max_compounds and (writer.NumMols() >= max_compounds):
                n_files += 1
                target = utils.prevent_overwrite(
                    os.path.join(dir_path, mine_db.name) + "_%s.sdf" % n_files)
                writer = AllChem.SmilesWriter(target)
    writer.close()


def export_smiles(mine_db, dir_path, max_compounds=None):
    """Exports compounds from the database as a SMILES file

    :param mine_db: a MINE object
    :param dir_path: directory for files
    :param max_compounds: maximum number of compounds per file (defaults to
        unlimited)
    :return:
    """
    header = ['SMILES', "_id", "Generation", 'Reactant_in', 'Product_of']
    # Make sure that all compounds point to all their reactants
    if not mine_db.compounds.find_one({"Product_of": {'$exists': 1}}):
        mine_db.add_rxn_pointers()

    print("Exporting %s compounds from %s as SMILES file" %
          (mine_db.compounds.count(), mine_db.name))
    target = open(utils.prevent_overwrite(os.path.join(dir_path, mine_db.name)
                                          + "_1.smiles"), 'w')

    # DictWriter allows for each key:value pair of a dictionary to be written
    # on its own row (by writerow)
    writer = csv.DictWriter(target, fieldnames=header, dialect='excel-tab')
    n_files = 1
    i = 0
    for compound in mine_db.compounds.find({}, dict([(x, 1) for x in header])):
        writer.writerow(compound)
        i += 1
        # If max compounds per file has been set by user and our number of
        # compounds that we have written so far is divisible by the max number,
        # then we start a new file
        if max_compounds and not i % max_compounds:
            n_files += 1
            target = open(
                utils.prevent_overwrite(os.path.join(dir_path, mine_db.name)
                                        + "_%s.smiles" % n_files), 'w')
            writer = csv.DictWriter(target, fieldnames=header,
                                    dialect='excel-tab')


def export_mol(mine_db, target, name_field='_id'):
    """Exports compounds from the database as MDL molfiles

    :param mine_db: The database to export
    :type mine_db: a MINE object
    :param target: a directory in which to place the files
    :type target: str
    :param name_field: the field to provide names for the mol files. Must be
        unique & universal
    :type name_field: str
    :return:
    :rtype:
    """
    # Create the file if it doesn't yet exist
    if not os.path.exists(target):
        os.mkdir(target)

    # Let user know if an id does not exist for every compound in database
    if mine_db.compounds.find().count() != mine_db.compounds.find(
            {name_field: {'$exists': 1}}).count():
        raise ValueError('%s does not exist for every compound in the database'
                         % name_field)

    for compound in mine_db.compounds.find({'_id': {'$regex': '^C'}}):
        # Create Mol object from SMILES code for each compound using
        # MolFromSmiles (rdkit). Take stereochemistry into account (True),
        # and replace CoA and R with *.
        mol = AllChem.MolFromSmiles(compound['SMILES'], True,
                                    {'CoA': '*', 'R': "*"})
        if "." in name_field:
            compound[name_field] = utils.get_dotted_field(compound, name_field)
        # Make things more compact and look nicer
        if isinstance(compound[name_field], list):
            compound[name_field] = ','.join(compound[name_field])
        # Use MolToMolFile (rdkit) to create a mol file from the Mol object
        # with the file path specified.
        AllChem.MolToMolFile(mol, os.path.join(target,
                                               compound[name_field] + '.mol'))


def export_tsv(mine_db, target, compound_fields=('_id', 'Names', 'Model_SEED',
                                                 'Formula', 'Charge', 'Inchi'),
               reaction_fields=('_id', 'SMILES_rxn', 'C_id_rxn')):
    """Exports MINE compound and reaction data as tab-separated values files
    amenable to use in ModelSEED.

    :param mine_db: The database to export
    :type mine_db: a MINE object
    :param target: a directory in which to place the files
    :type target: string
    :param compound_fields: The fields to export in the compound table
    :type compound_fields: set
    :param reaction_fields: The fields to export in the reaction table
    :type reaction_fields: set
    """
    db_links = ('KEGG', 'Model_SEED', 'PubChem')
    print("Exporting %s compounds from %s to tsv" % (mine_db.compounds.count(),
                                                     mine_db.name))
    with open(utils.prevent_overwrite(os.path.join(target, mine_db.name)
                                      + "_compounds.tsv"), 'w') as out:
        writer = csv.DictWriter(out, fieldnames=compound_fields,
                                dialect='excel-tab')
        writer.writeheader()
        for compound in mine_db.compounds.find(
                {}, dict([('SMILES', 1)] + [('DB_links.' + x, 1) if x
                                            in db_links else (x, 1) for x
                                            in compound_fields])):
            # This is a work around for supporting older MINEs which lack Inchi
            if 'Inchi' in compound_fields and 'Inchi' not in compound:
                compound['Inchi'] = AllChem.MolToInchi(AllChem.MolFromSmiles(
                    compound['SMILES']))
            if 'SMILES' not in compound_fields:
                del compound['SMILES']
            if 'DB_links' in compound:
                for k, v in compound['DB_links'].items():
                    compound[k] = ", ".join(v)
                del compound['DB_links']
            writer.writerow(compound)

    print("Exporting %s reactions from %s to tsv" % (mine_db.reactions.count(),
                                                     mine_db.name))
    with open(utils.prevent_overwrite(os.path.join(target, mine_db.name)
                                      + "_reactions.tsv"), 'w') as out:
        writer = csv.DictWriter(out, fieldnames=reaction_fields,
                                dialect='excel-tab')
        writer.writeheader()
        for rxn in mine_db.reactions.find(
                {}, dict([('Reactants', 1), ('Products', 1)]
                         + [(x, 1) for x in reaction_fields])):
            if 'C_id_rxn' in reaction_fields:
                def to_str(half_rxn):
                    return ['(%s) %s' % (x['stoich'], x['c_id'])
                            for x in half_rxn]
                rxn['C_id_rxn'] = ' + '.join(to_str(rxn['Reactants'])) + \
                                  ' => ' + ' + '.join(to_str(rxn['Products']))
            if 'Reactants' not in reaction_fields:
                del rxn['Reactants']
            if 'Products' not in reaction_fields:
                del rxn['Products']
            writer.writerow(rxn)


def export_kbase(mine_db, target):
    """Exports MINE compound and reaction data as tab-separated values files
    amenable to use in ModelSEED.

    :param mine_db: The database to export
    :type mine_db: a MINE object
    :param target: a directory in which to place the files
    :type target: str
    """
    compound_fields = collections.OrderedDict([('id', "_id"), ('name', ""),
                                               ('formula', 'Formula'),
                                               ('charge', 'Charge'),
                                               ('aliases', "Names")])
    reaction_fields = collections.OrderedDict(
        [('id', "_id"), ('direction', ">"), ('compartment', "c0"), ('gpr', ''),
         ('name', ''), ('enzyme', ''), ('pathway', ''), ('reference', ''),
         ('equation', '')])
    print("Exporting %s compounds from %s to tsv" % (mine_db.compounds.count(),
                                                     mine_db.name))
    with open(utils.prevent_overwrite(os.path.join(target, mine_db.name)
                                      + "_compounds.tsv"), 'w') as out:
        writer = csv.DictWriter(out, fieldnames=compound_fields,
                                dialect='excel-tab')
        writer.writeheader()
        for compound in mine_db.compounds.find(
                {}, dict([("Names", 1), ('DB_links.Model_SEED', 1)]
                         + [(x, 1) for x in compound_fields.values()])):
            if compound['_id'][0] == 'X':
                continue
            for k, v in compound_fields.items():
                if v in compound:
                    compound[k] = compound[v]
                    del compound[v]
            if 'name' in compound_fields and 'Names' in compound:
                compound['name'] = compound['Names'][0]
                del compound['Names']
            if 'aliases' in compound:
                compound['aliases'] = "|".join(compound['aliases'])
                if 'Model_SEED' in compound['DB_links']:
                    compound['aliases'] += "|" + "|".join(sorted(
                        compound['DB_links']['Model_SEED']))
            if 'DB_links' in compound:
                del compound['DB_links']
            writer.writerow(compound)

    print("Exporting %s reactions from %s to tsv" % (mine_db.reactions.count(),
                                                     mine_db.name))
    with open(utils.prevent_overwrite(os.path.join(target, mine_db.name)
                                      + "_reactions.tsv"), 'w') as out:
        writer = csv.DictWriter(out, fieldnames=reaction_fields,
                                dialect='excel-tab')
        writer.writeheader()
        for rxn in mine_db.reactions.find(
                {}, dict([('Reactants', 1), ('Products', 1)]
                         + [(x, 1) for x in reaction_fields.values()])):
            for k, v in reaction_fields.items():
                if v in rxn:
                    rxn[k] = rxn[v]
                    del rxn[v]
            if 'equation' in reaction_fields:
                def to_str(half_rxn):
                    return ['(%s) %s' % (x['stoich'],
                                         x['c_id'].replace('X', 'C'))
                            for x in half_rxn]
                rxn['equation'] = ' + '.join(to_str(rxn['Reactants'])) + \
                                  ' => ' + ' + '.join(to_str(rxn['Products']))
            if 'Reactants' not in reaction_fields:
                del rxn['Reactants']
            if 'Products' not in reaction_fields:
                del rxn['Products']
            writer.writerow(rxn)


def export_inchi_rxns(mine_db, target, rxn_ids=None):
    """Export reactions from a MINE db to a .tsv file.

    :param mine_db: name of MongoDB to export reactions from
    :param target: path to folder to save .tsv export file in
    :param rxn_ids: only export reactions with these ids (list)
    """
    reaction_fields = collections.OrderedDict(
        [("Reaction Rule", "Operators"), ('ID', "_id"), ('Equation', '')])
    comp_memo = {}

    def get_name_and_inchi(comp_id):
        if comp_id not in comp_memo:
            comp = mine_db.compounds.find_one({"_id": comp_id},
                                              {"Names": 1, "Inchi": 1,
                                               "MINE_id": 1})
            comp_memo[comp_id] = (comp.get('Names', [comp['MINE_id']])[0],
                                  comp.get('Inchi'))
        return comp_memo[comp_id]

    def to_str(half_rxn):
        lst = []
        for x in half_rxn:
            name, inchi = get_name_and_inchi(x['c_id'])
            lst.append('(%s) %s[%s]' % (x['stoich'], name, inchi))
        return lst

    with open(utils.prevent_overwrite(os.path.join(target, mine_db.name)
                                      + "_reactions.tsv"), 'w') as out:
        writer = csv.DictWriter(out, fieldnames=reaction_fields,
                                dialect='excel-tab')
        writer.writeheader()
        if rxn_ids:
            query = {"_id": {"$in": rxn_ids}}
        else:
            query = {}
        for rxn in mine_db.reactions.find(
                query, dict([('Reactants', 1), ('Products', 1)]
                            + [(x, 1) for x in reaction_fields.values()])):
            for k, v in reaction_fields.items():
                if v in rxn:
                    if isinstance(rxn[v], list):
                        rxn[k] = ", ".join(rxn[v])
                    else:
                        rxn[k] = rxn[v]
                    del rxn[v]
            if 'Equation' in reaction_fields:
                rxn['Equation'] = ' + '.join(to_str(rxn['Reactants'])) + \
                                  ' => ' + ' + '.join(to_str(rxn['Products']))
            if 'Reactants' not in reaction_fields:
                del rxn['Reactants']
            if 'Products' not in reaction_fields:
                del rxn['Products']
            writer.writerow(rxn)


def import_sdf(mine_db, target):
    """Imports a SDF file as a MINE database

    :param mine_db: a MINE object, the database to insert the compound into
    :param target: a path, the SDF file to be loaded
    :return:
    """
    # SDMolSupplier (rdkit) takes entries from sdf file and returns Mol objects
    sdf_gen = AllChem.SDMolSupplier(target)
    # Go through each generated Mol object and add each to MINE database
    for mol in sdf_gen:
        mine_db.insert_compound(mol, compound_dict=mol.GetPropsAsDict(),
                                pubchem_db=None, kegg_db=None,
                                modelseed_db=None)
    # Add to log file (metadata)
    mine_db.meta_data.insert({"Timestamp": datetime.datetime.now(), "Action":
                              "SDF Imported", "Filepath": target})


def import_smiles(mine_db, target):
    """
    Imports a smiles file as a MINE database
    :param mine_db: a MINE object, the database to insert the compound into
    :param target: a path, the SDF file to be loaded
    :return:
    """
    # SmilesMolSupplier (rdkit) generates Mol objects from smiles file (.smi)
    mols = AllChem.SmilesMolSupplier(target, delimiter='\t', nameColumn=0)
    # Go through each generated mol file and add molecule to MINE database
    # Stores compound properties in dict (GetPropsAsDict() from rdkit Mol
    # class)
    for mol in mols:
        if mol:
            mine_db.insert_compound(mol, compound_dict=mol.GetPropsAsDict(),
                                    pubchem_db=None, kegg_db=None,
                                    modelseed_db=None)
    # Add to log file (metadata)
    mine_db.meta_data.insert({"Timestamp": datetime.datetime.now(), "Action":
                              "SDF Imported", "Filepath": target})


def import_mol_dir(mine_db, target, name_field="Name", overwrite=False):
    """Imports a directory of molfiles as a MINE database

    :param mine_db: a MINE object, the database to insert the compound into
    :param target: a path, the molfile directory to be loaded
    :param name_field: a string, the field for the compound name
    :param overwrite: a bool, if true, new compounds replace the old compounds
        in the database
    """
    # For each .mol file in the directory of the target folder (path):
    for file in os.listdir(target):
        if ".mol" in file:
            # MolFromMolFile (rdkit) generates Mol objects from .mol files
            mol = AllChem.MolFromMolFile(target + '/' + file)
            # Mol object name becomes name of mol file without .mol extension
            name = file.rstrip('.mol')
            # Check that Mol object is successfully generated
            if mol:
                # Create hashkey for the compound
                cpdhash = utils.compound_hash(mol)
                # If we don't want to overwrite, and the compound (cpdhash)
                # already exists, then add an extra cpdhash for that molecule
                if not overwrite and mine_db.compounds.count({"_id": cpdhash}):
                    mine_db.compounds.update({"_id": cpdhash},
                                             {"$addToSet": {name_field: name}})
                # If we don't care about overwriting, just insert the new
                # compound into the database
                else:
                    mine_db.insert_compound(mol,
                                            compound_dict={name_field: [name],
                                                           'Generation': 0},
                                            pubchem_db=None, kegg_db=None,
                                            modelseed_db=None)
    # Add to log file (metadata)
    mine_db.meta_data.insert({"Timestamp": datetime.datetime.now(), "Action":
                              "MolFiles Imported", "Filepath": target})


if __name__ == '__main__':
    # User inputs task as first argument (export-sdf, export-smi, export-mol,
    #  import-sdf, import-smi, or import-mol)
    TASK = sys.argv[1]
    # User inputs database name as second argument
    DB_NAME = sys.argv[2]
    # User inputs file path as third argument
    PATH = sys.argv[3]
    database = MINE(DB_NAME)  # pylint: disable=invalid-name
    if TASK == 'export-sdf':
        # If a maximum molecules per file is specified (fourth argument
        # entered by user), then pass that to the export function.
        if len(sys.argv) == 5:
            export_sdf(database, PATH, int(sys.argv[4]))
        # Otherwise, assume an unlimited number of molecules per file
        else:
            export_sdf(database, PATH)
    elif TASK == 'export-smi':
        # If a maximum molecules per file is specified (fourth argument
        # entered by user), then pass that to the export function.
        if len(sys.argv) == 5:
            export_smiles(database, PATH, int(sys.argv[4]))
        # Otherwise, assume an unlimited number of molecules per file
        else:
            export_smiles(database, PATH)
    elif TASK == 'export-mol':
        # If a maximum molecules per file is specified (fourth argument
        # entered by user), then pass that to the export function.
        if len(sys.argv) == 5:
            export_mol(database, PATH, sys.argv[4])
        # Otherwise, assume an unlimited number of molecules per file
        else:
            export_mol(database, PATH)
    elif TASK == 'export-tsv':
        export_tsv(database, PATH)
    elif TASK == 'export-kbase':
        export_kbase(database, PATH)
    elif TASK == 'import-sdf':
        import_sdf(database, PATH)
    elif TASK == 'import-smi':
        import_smiles(database, PATH)
    elif TASK == 'import-mol':
        import_mol_dir(database, PATH)
    else:
        print("ERROR: Unrecognised TASK")



"""Tests for databases.py using pytest."""
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

import json
import os
import unittest
from shutil import rmtree

import pymongo
import pytest
from pymongo.errors import ServerSelectionTimeoutError
from rdkit.Chem import AllChem

from minedatabase import utils

from minedatabase.databases import MINE


@pytest.fixture()
def test_db():
    """Create a test MINE database. Created and torn down before and after each
    test it is used in."""
    print(os.path.dirname(__file__))
    datafile_path = os.path.join(os.path.dirname(__file__),
                                 'data/testing_db.json')

    try:
        testdb = MINE("mongotest")
        with open(datafile_path) as infile:
            jsondb = json.load(infile)
        for doc in jsondb[0]:
            if testdb.compounds.find_one({'_id': doc['_id']}):
                testdb.compounds.replace_one({'_id': doc['_id']}, doc)
            else:
                testdb.compounds.insert_one(doc)
        for doc in jsondb[1]:
            if testdb.reactions.find_one({'_id': doc['_id']}):
                testdb.reactions.replace_one({'_id': doc['_id']}, doc)
            else:
                testdb.reactions.insert_one(doc)
        for doc in jsondb[2]:
            if testdb.operators.find_one({'_id': doc['_id']}):
                testdb.operators.replace_one({'_id': doc['_id']}, doc)
            else:
                testdb.operators.insert_one(doc)

    except ServerSelectionTimeoutError:
        print('No Mongo DB server detected')

    yield testdb


@unittest.skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                 "Skipping this test on Travis CI.")
def test_generate_image_files(test_db):
    """
    GIVEN a MINE database
    WHEN image files are generated from that database
    THEN make sure that they are all generated correctly
    """
    img_dir = os.path.join(os.path.dirname(__file__), 'imgs')
    test_db.generate_image_files(img_dir)
    try:
        assert os.path.exists(
            os.path.join(img_dir,
                         './C455bc3dc93cd3bb3bef92a34767693a4716aa3fb.svg'))
        assert len(os.listdir(img_dir)) == 26
    finally:
        rmtree(img_dir)
    # Use 'Generation': 1 to do this for just the starting compound
    test_db.generate_image_files(img_dir, {'Generation': 1}, 3, 'png')
    try:
        assert os.path.exists(
            os.path.join(img_dir, 'C', 'c', 'f',
                         'Ccffda1b2e82fcdb0e1e710cad4d5f70df7a5d74f.png'))
    finally:
        rmtree(img_dir)


def test_insert_compound(test_db):
    """
    GIVEN a compound (Mol object) with associated properties
    WHEN that compound and its properties are added to a MINE database
    THEN check that the compound and its properties are correctly stored
    """
    smiles = 'CC(=O)O'
    mol = AllChem.MolFromSmiles(smiles)
    test_db.insert_compound(mol, {'Generation': 0.0})
    try:
        entry = test_db.compounds.find_one({"SMILES": smiles})
        assert entry
        assert isinstance(entry['Mass'], float)
        assert len(entry['RDKit']) > 0
        assert len(entry['RDKit']) == entry['len_RDKit']
        assert entry["NP_likeness"]
        assert entry['logP']
    finally:
        test_db.compounds.delete_many({"SMILES": smiles})


def test_insert_reaction(test_db):
    """
    GIVEN a reaction (dict with 'Equation', 'Reactants', and 'Products' keys)
    WHEN that reaction is inserted into a MINE database
    THEN check that it is present and that the stoichiometry is correct
    """
    rxn = {'Equation': 'A + B = C + D'}
    rxn['Reactants'], rxn['Products'] = utils.parse_text_rxn(rxn['Equation'],
                                                             ' = ', ' + ')
    test_db.insert_reaction(rxn)
    entry = test_db.reactions.find_one({"_id": "4542c96f4bca04bfe2db15bc71e9"
                                               "eaee38bee5b87ad8a6752a5c4718"
                                               "ba1974c1"})
    assert entry
    assert isinstance(entry['_id'], str) and (len(entry['_id']) > 0)
    assert entry['Reactants'] == [{"stoich": 1, "c_id": "A"},
                                  {"stoich": 1, "c_id": "B"}]
    assert entry['Products'] == [{"stoich": 1, "c_id": "C"},
                                 {"stoich": 1, "c_id": "D"}]


def test_init(test_db):
    """
    GIVEN a MINE database
    WHEN N/A
    THEN make sure that all database field types are correct
    """
    assert isinstance(test_db.compounds, pymongo.collection.Collection)
    assert isinstance(test_db.reactions, pymongo.collection.Collection)
    assert isinstance(test_db.operators, pymongo.collection.Collection)
    assert isinstance(test_db._db, pymongo.database.Database)
    assert isinstance(test_db.id_db, pymongo.database.Database)



"""Tests for pickaxe.py using pytest."""
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

import os
import re
import hashlib
import subprocess
from filecmp import cmp

import pytest

from rdkit.Chem import AllChem
from minedatabase import pickaxe
from minedatabase.databases import MINE

DATA_DIR = os.path.dirname(__file__) + '/data'


@pytest.fixture
def smiles_dict():
    """Store SMILES for compounds used in test cases here."""
    smiles = {
        'ATP': 'Nc1ncnc2c1ncn2[C@@H]1O[C@H](COP(=O)(O)OP(=O)(O)OP(=O)(O)O)[C'
               + '@@H](O)[C@H]1O',
        'ADP': 'Nc1ncnc2c1ncn2[C@@H]1O[C@H](COP(=O)(O)OP(=O)(O)O)[C@@H](O)[C'
               + '@H]1O',
        'meh': 'CCC(=O)C(=O)O',
        'l_ala': 'C[C@H](N)C(=O)O',
        'd_ala': 'C[C@@H](N)C(=O)O',
        'FADH': 'Cc1cc2c(cc1C)N(CC(O)C(O)C(O)COP(=O)(O)OP(=O)(O)OCC1OC(n3cnc'
                + '4c(N)ncnc43)C(O)C1O)c1nc(O)nc(O)c1N2',
        'S-Adenosylmethionine': 'C[S+](CC[C@H](N)C(=O)O)C[C@H]1O[C@@H](n2cnc'
                                + '3c(N)ncnc32)[C@H](O)[C@@H]1O'
    }
    return smiles


@pytest.fixture
def coreactant_dict(smiles_dict):
    """Create tab-formatted coreactant entries."""
    coreactants = {
        'ATP': 'ATP		' + smiles_dict['ATP'],
        'ADP': 'ADP		' + smiles_dict['ADP'],
        'S-Adenosylmethionine': 'S-Adenosylmethionine		'
                                + smiles_dict['S-Adenosylmethionine']
    }
    return coreactants


@pytest.fixture
def pk():
    """Create default Pickaxe object."""
    return pickaxe.Pickaxe(coreactant_list=DATA_DIR + '/test_coreactants.tsv',
                           rule_list=DATA_DIR + '/test_reaction_rules.tsv')


@pytest.fixture
def default_rule(pk):
    """Default operator to use for testing is set here."""
    return pk.rxn_rules['2.7.1.a']


@pytest.fixture
def pk_transformed(default_rule, smiles_dict, coreactant_dict):
    """Create Pickaxe object with a few predicted reactions."""
    pk_transformed = pickaxe.Pickaxe()
    pk_transformed._add_compound("Start", smi=smiles_dict['FADH'])
    pk_transformed._load_coreactant(coreactant_dict['ATP'])
    pk_transformed._load_coreactant(coreactant_dict['ADP'])
    pk_transformed.rxn_rules['2.7.1.a'] = default_rule
    pk_transformed.transform_compound(smiles_dict['FADH'])
    pk_transformed.assign_ids()
    return pk_transformed


def purge(directory, pattern):
    """Delete all files in a directory matching a regex pattern."""
    for filename in os.listdir(directory):
        if re.search(pattern, filename):
            os.remove(os.path.join(directory, filename))


def multiprocess(pk, smiles_dict, coreactant_dict):
    """Convert FADH into other compounds in parallel."""
    pk._load_coreactant(coreactant_dict['ATP'])
    pk._load_coreactant(coreactant_dict['ADP'])
    pk._add_compound(smiles_dict['FADH'], smiles_dict['FADH'],
                     cpd_type='Starting Compound')
    pk.transform_all(max_generations=2, num_workers=2)
    return pk


def test_cofactor_loading(pk):
    """
    GIVEN a default Pickaxe object
    WHEN cofactors are loaded into the Pickaxe object in its creation
    THEN make sure those cofactors were loaded correctly
    """
    assert "O=C=O" in pk._raw_compounds
    c_id = pk._raw_compounds['O=C=O']
    assert c_id in pk.compounds
    assert pk.compounds[c_id]['Type'] == 'Coreactant'
    assert isinstance(pk.coreactants['ATP'][0], AllChem.Mol)
    assert pk.coreactants['ATP'][1][0] == "X"


def test_reaction_rule_loading(default_rule):
    """
    GIVEN a reaction rule dict
    WHEN reaction rules are loaded during Pickaxe object initialization
    THEN make sure it is formatted correctly
    """
    assert isinstance(default_rule[0], AllChem.ChemicalReaction)
    assert isinstance(default_rule[1], dict)
    assert default_rule[1]['Reactants'] == ['ATP', 'Any']
    assert "Products" in default_rule[1]
    assert "Comments" in default_rule[1]


def test_compound_loading(pk):
    """
    GIVEN a default Pickaxe object
    WHEN compounds are loaded
    THEN check that they are loaded correctly
    """
    compound_smiles = pk.load_compound_set(
        compound_file=DATA_DIR + '/test_compounds.tsv')
    assert len(compound_smiles) == 14


def test_transform_compounds_implicit(smiles_dict):
    """
    GIVEN a compound (meh in this case)
    WHEN that compound is transformed into another compound via pickaxe
    THEN make sure that the transformation is successful and recorded
    """
    pk = pickaxe.Pickaxe(explicit_h=False, kekulize=False,
                         coreactant_list=DATA_DIR + '/test_coreactants.tsv',
                         rule_list=DATA_DIR + '/test_cd_rxn_rule.tsv')
    pk._add_compound("Start", smi=smiles_dict['meh'])
    pk.transform_compound(smiles_dict['meh'])
    assert len(pk.compounds) == 38
    assert len(pk.reactions) == 1


def test_hashing(pk, smiles_dict, coreactant_dict):
    """
    GIVEN a default Pickaxe object
    WHEN compounds with very similar structures are both transformed
    THEN make sure they result in different products (hashed differently)
    """
    pk._load_coreactant(coreactant_dict['S-Adenosylmethionine'])
    pk.transform_compound(smiles_dict['l_ala'])
    len_rxns = len(pk.reactions)
    assert len_rxns > 0
    pk.transform_compound(smiles_dict['d_ala'])
    # If hashed differently, should now have an extra set of rxns for d_ala
    assert len(pk.reactions) == 2 * len_rxns


def test_product_racimization(smiles_dict):
    """
    GIVEN molecules of undefined steriochemistry
    WHEN transforming them via pickaxe with or without racemization
    THEN make sure that when racemized, we observe extra reaction(s)
    """
    pk = pickaxe.Pickaxe(racemize=False,
                         coreactant_list=DATA_DIR + '/test_coreactants.tsv',
                         rule_list=DATA_DIR + '/test_reaction_rules.tsv')
    comps, rxns = pk.transform_compound(smiles_dict['meh'], rules=['2.6.1.a'])
    assert len(comps) == 38
    assert len(rxns) == 1

    pk = pickaxe.Pickaxe(racemize=True,
                         coreactant_list=DATA_DIR + '/test_coreactants.tsv',
                         rule_list=DATA_DIR + '/test_reaction_rules.tsv')
    rcomps, rrxns = pk.transform_compound(smiles_dict['meh'],
                                          rules=['2.6.1.a'])
    # Extra reaction occurs due to undefined chiral center giving rise to
    # two possible starting structures in this second case
    assert len(rcomps) == 39
    assert len(rrxns) == 2


def test_compound_output_writing(pk_transformed):
    """
    GIVEN a Pickaxe object with predicted transformations
    WHEN all compounds (including predicted) are written to an output file
    THEN make sure they are correctly written, and that they are all present
    """
    with open(DATA_DIR + '/testcompoundsout.tsv', 'rb') as infile:
        expected = hashlib.sha256(infile.read()).hexdigest()
    pk_transformed.write_compound_output_file(DATA_DIR
                                              + '/testcompoundsout.tsv')
    assert os.path.exists(DATA_DIR + '/testcompoundsout_new.tsv')
    try:
        with open(DATA_DIR + '/testcompoundsout_new.tsv', 'rb') as infile:
            output_compounds = hashlib.sha256(infile.read()).hexdigest()
        assert expected == output_compounds
    finally:
        os.remove(DATA_DIR + '/testcompoundsout_new.tsv')


def test_reaction_output_writing(pk_transformed):
    """
    GIVEN a Pickaxe object with predicted transformations
    WHEN all reactions (including predicted) are written to an output file
    THEN make sure they are correctly written, and that they are all present
    """
    with open(DATA_DIR + '/testreactionsout.tsv', 'rb') as infile:
        expected = hashlib.sha256(infile.read()).hexdigest()
    pk_transformed.write_reaction_output_file(DATA_DIR
                                              + '/testreactionsout.tsv')
    assert os.path.exists(DATA_DIR + '/testreactionsout_new.tsv')
    try:
        with open(DATA_DIR + '/testreactionsout_new.tsv', 'rb') as infile:
            output_compounds = hashlib.sha256(infile.read()).hexdigest()
        assert expected == output_compounds
    finally:
        os.remove(DATA_DIR + '/testreactionsout_new.tsv')


def test_transform_all(default_rule, smiles_dict, coreactant_dict):
    """
    GIVEN a set of rules and starting compounds
    WHEN we run pickaxe to predict potential transformations
    THEN make sure all expected transformations are predicted
    """
    pk = pickaxe.Pickaxe(errors=False)
    pk._load_coreactant(coreactant_dict['ATP'])
    pk._load_coreactant(coreactant_dict['ADP'])
    pk._add_compound(smiles_dict['FADH'], smiles_dict['FADH'],
                     cpd_type='Starting Compound')
    pk.rxn_rules['2.7.1.a'] = default_rule
    pk.transform_all(max_generations=2)
    assert len(pk.compounds) == 31
    assert len(pk.reactions) == 49
    comp_gens = set([x['Generation'] for x in pk.compounds.values()])
    assert comp_gens == {0, 1, 2}


def test_multiprocessing(pk, smiles_dict, coreactant_dict):
    """
    GIVEN a Pickaxe object
    WHEN we use multiprocessing to enumerate predicted reactions
    THEN make sure those predictions are correct
    """
    pk = multiprocess(pk, smiles_dict, coreactant_dict)
    assert len(pk.compounds) == 67
    assert len(pk.reactions) == 49
    comp_gens = set([x['Generation'] for x in pk.compounds.values()])
    assert comp_gens == {0, 1, 2}


def test_cli():
    """
    GIVEN the pickaxe CLI
    WHEN pickaxe is run from the command line
    THEN make sure it exits with exit code 0 (no errors)
    """
    os.chdir(DATA_DIR + "/../..")
    rc = subprocess.call(
        'python minedatabase/pickaxe.py -o tests -r '
        'tests/data/test_cd_rxn_rule.tsv',
        shell=True)
    assert not rc
    purge('tests/', r".*\.tsv$")


def test_pruning(default_rule, smiles_dict, coreactant_dict):
    """
    GIVEN a Pickaxe expansion
    WHEN that expansion is pruned via Pickaxe.prune_network()
    THEN make sure that the pruned compounds no longer exist in the network
    """
    pk = pickaxe.Pickaxe(database='MINE_test', image_dir=DATA_DIR)
    pk.rxn_rules['2.7.1.a'] = default_rule
    pk = multiprocess(pk, smiles_dict, coreactant_dict)
    ids = ['C9d4089d24d09c0c86a817af75ded95dd0f6d5b07',
           'C177697ec7f877acf4f1439ce57c03c22fbe5f897', 'C69']
    pk.prune_network(ids)
    pk.assign_ids()
    pk.write_compound_output_file(DATA_DIR + '/pruned_comps')
    pk.write_reaction_output_file(DATA_DIR + '/pruned_rxns')
    assert os.path.exists(DATA_DIR + '/pruned_comps_new')
    assert os.path.exists(DATA_DIR + '/pruned_rxns_new')
    try:
        assert cmp(DATA_DIR + '/pruned_comps', DATA_DIR + '/pruned_comps_new')
    finally:
        os.remove(DATA_DIR + '/pruned_comps_new')
    try:
        assert cmp(DATA_DIR + '/pruned_rxns', DATA_DIR + '/pruned_rxns_new')
    finally:
        os.remove(DATA_DIR + '/pruned_rxns_new')


def test_load_compounds_from_mine():
    """
    GIVEN a Pickaxe object loaded from the 'mongotest' MINE DB
    WHEN Pickaxe.load_compound_set() is used to load all compounds from the DB
    THEN make sure that all compounds are loaded
    """
    pk = pickaxe.Pickaxe(database='mongotest')
    compound_smiles = pk.load_compound_set()
    assert len(compound_smiles) == 26


def test_save_as_mine(default_rule, smiles_dict, coreactant_dict):
    """
    GIVEN a Pickaxe expansion
    WHEN that expansion is saved as a MINE DB in the MongoDB
    THEN make sure that all features are saved in the MongoDB as expected
    """
    pk = pickaxe.Pickaxe(database='MINE_test', image_dir=DATA_DIR)
    pk.rxn_rules['2.7.1.a'] = default_rule
    pk = multiprocess(pk, smiles_dict, coreactant_dict)
    pk.save_to_mine("MINE_test")
    mine_db = MINE('MINE_test')
    try:
        assert mine_db.compounds.estimated_document_count() == 31
        assert mine_db.reactions.estimated_document_count() == 49
        assert mine_db.operators.estimated_document_count() == 1
        assert mine_db.operators.find_one()["Reactions_predicted"] == 49
        assert os.path.exists(
            DATA_DIR + '/X9c69cbeb40f083118c1913599c12c7f4e5e68d03.svg')
        start_comp = mine_db.compounds.find_one({'Type': 'Starting Compound'})
        assert len(start_comp['Reactant_in']) > 0
        # Don't track sources of coreactants
        coreactant = mine_db.compounds.find_one({'Type': 'Coreactant'})
        assert 'Sources' not in coreactant
        product = mine_db.compounds.find_one({'Generation': 2})
        assert len(product['Product_of']) > 0
        assert len(product['Sources']) > 0
        assert product['Type'] == 'Predicted'
    finally:
        mine_db.compounds.drop()
        mine_db.reactions.drop()
        mine_db.operators.drop()
        purge(DATA_DIR, r".*\.svg$")


def test_save_no_rxn_mine():
    """
    GIVEN a Pickaxe object with no expansion
    WHEN that Pickaxe object is saved into a MINE DB in the MongoDB
    THEN check that starting compounds are present and that no reactions exist
    """
    pk = pickaxe.Pickaxe(database='MINE_test')
    pk.load_compound_set(compound_file=DATA_DIR + '/test_compounds.tsv')
    pk.save_to_mine("MINE_test")
    mine_db = MINE('MINE_test')
    try:
        assert mine_db.compounds.estimated_document_count() == 14
        assert mine_db.reactions.estimated_document_count() == 0
    finally:
        mine_db.compounds.drop()
        mine_db.reactions.drop()
        mine_db.operators.drop()



"""Tests for queries.py using pytest."""
# pylint: disable=redefined-outer-name

import json
from os.path import dirname

import pytest
from pymongo.errors import ServerSelectionTimeoutError

from minedatabase import databases, queries

from minedatabase.databases import MINE


@pytest.fixture()
def test_db():
    """Create a test MINE database. Created and torn down before and after each
    test it is used in."""
    try:
        testdb = MINE("mongotest")
    except ServerSelectionTimeoutError:
        print('No Mongo DB server detected')
    yield testdb


@pytest.fixture()
def test_molfile():
    """Mol file for glucose compound."""
    test_molfile = open(dirname(__file__) + "/data/glucose.mol", "r").read()
    return test_molfile


@pytest.fixture()
def glucose():
    """MongoDB document (.json) for glucose compound."""
    with open(dirname(__file__) + '/data/glucose.json') as infile:
        glucose = json.load(infile)
    return glucose


@pytest.fixture
def glucose_id():
    """ID in MongoDB for glucose."""
    glucose_id = {'_id': 'Ccffda1b2e82fcdb0e1e710cad4d5f70df7a5d74f'}
    return glucose_id


def test_quick_search(test_db, glucose, glucose_id):
    """
    GIVEN a quick search query (e.g. glucose identifiers)
    WHEN quick search is used to search based on that query
    THEN make sure that quick search provides the correct results
    """
    assert glucose not in queries.quick_search(test_db,
                                               'WQZGKKKJIJFFOK-UHFFFAOYSA-N')
    assert glucose in queries.quick_search(
        test_db, 'InChIKey=WQZGKKKJIJFFOK-GASJEMHNSA-N')
    assert glucose in queries.quick_search(
        test_db, "Ccffda1b2e82fcdb0e1e710cad4d5f70df7a5d74f")
    assert glucose in queries.quick_search(test_db, "917030")
    assert glucose in queries.quick_search(test_db, "cpd00027")
    assert glucose in queries.quick_search(test_db, 'C00031')
    assert glucose in queries.quick_search(test_db, 'Glucose')
    assert glucose_id in queries.quick_search(
        test_db, 'WQZGKKKJIJFFOK-GASJEMHNSA-N', {'_id': 1})


def test_database_query(test_db, glucose, glucose_id):
    """
    GIVEN an andvanced search query (e.g. a MINE id)
    WHEN advanced search is used to search based on that query
    THEN make sure that advanced search provides the correct results
    """
    with pytest.raises(ValueError):
        queries.advanced_search(databases.MINE('admin'), "{'MINE_id': 19160}")
    with pytest.raises(ValueError):
        queries.advanced_search(test_db, "")
    assert queries.advanced_search(test_db, "{'MINE_id': 917030}") == [glucose]
    assert queries.advanced_search(test_db,
                                   "{'Names': 'Glucose'}") == [glucose]
    assert queries.advanced_search(test_db, "{'MINE_id': 917030}",
                                   {'_id': 1}) == [glucose_id]


def test_similarity_search(test_db, test_molfile, glucose):
    """
    GIVEN a similary search query
    WHEN similarity search is used to search based on that query
    THEN make sure the similarity search provides the correct results
    """
    assert len(queries.similarity_search(
        test_db, 'Cc1cc2c(cc1C)N(CC(O)C(O)C(O)COP(=O)(O)OP(=O)(O)OCC1OC(n3cn'
        'c4c(N)ncnc43)C(O)C1O)c1nc(O)nc(O)c1N2', 0.9, 100)) == 8
    result = queries.similarity_search(test_db, test_molfile, 0.5, 100,
                                       fp_type='MACCS')
    assert glucose in result
    assert len(result) == 3


def test_substructure_search(test_db, glucose):
    """
    GIVEN a substructure search query
    WHEN substructure search is used to search based on that query
    THEN make sure that substructure search provides the correct results
    """
    result = queries.substructure_search(test_db, "CO", 100)
    assert glucose in result
    assert len(result) == 22
    result = queries.substructure_search(test_db, 'O=P(O)(O)O', 100)
    assert len(result) == 15
    assert isinstance(result[0], dict)


def test_structure_search(test_db, test_molfile, glucose):
    """
    GIVEN a structure search query
    WHEN structure search is used to search based on that query
    THEN make sure that structure search provides the correct results
    """
    assert glucose in queries.structure_search(
        test_db, 'OC[C@H]1OC(O)[C@H](O)[C@@H](O)[C@@H]1O', True)
    assert glucose in queries.structure_search(test_db, test_molfile, False)



"""Utils.py: contains basic functions reused in various contexts in other
modules"""
import collections
import csv
import hashlib
import json
from os import path

from rdkit.Chem import AllChem

StoichTuple = collections.namedtuple("StoichTuple", 'stoich,c_id')


def file_to_dict_list(filepath):
    """Accept a path to a CSV, TSV or JSON file and return a dictionary list"""
    if '.tsv' in filepath:
        reader = csv.DictReader(open(filepath), dialect='excel-tab')
    elif '.csv' in filepath:
        reader = csv.DictReader(open(filepath))
    elif '.json' in filepath:
        reader = json.load(open(filepath))
    else:
        raise ValueError('Unrecognized input file type')
    return list(reader)


def compound_hash(compound, cofactor=False):
    """Creates hash string for given compound

    :param compound: The compound to be hashed
    :type compound: str or Mol Object
    :param cofactor: is the compound a cofactor
    :type cofactor: bool
    :return: A hashed compound _id
    :rtype: str
    """
    # Check to see if compound is a Mol object. If true, convert that Mol
    # object to a SMILES string
    if isinstance(compound, AllChem.Mol):
        compound = AllChem.MolToSmiles(compound, True)
    # Create hash string using hashlib module
    chash = hashlib.sha1(compound.encode('utf-8')).hexdigest()
    # Mark cofactors with an X at the beginning, all else with a C
    if cofactor:
        return "X" + chash
    else:
        return "C" + chash


def convert_sets_to_lists(obj):
    """Recursively converts dictionaries that contain sets to lists"""
    if isinstance(obj, set):
        # This brings short names to the top of the list
        try:
            obj = sorted(list(obj), key=len)
        except TypeError:
            obj = list(obj)
    elif isinstance(obj, dict):
        for key in obj:
            obj[key] = convert_sets_to_lists(obj[key])
    return obj


def get_dotted_field(input_dict, accessor_string):
    """Gets data from a dictionary using a dotted accessor-string

    :param input_dict: A nested dictionary
    :type input_dict: dict
    :param accessor_string: A dotted path description i.e. "DBLinks.KEGG"
    :type accessor_string: str
    :return: The value in the nested dict
    """
    current_data = input_dict
    for chunk in accessor_string.split('.'):
        current_data = current_data.get(chunk, {})
    return current_data


def save_dotted_field(accessor_string, data):
    """Saves data to a dictionary using a dotted accessor-string

    :param accessor_string: A dotted path description i.e. "DBLinks.KEGG"
    :type accessor_string: str
    :param data: The value to be stored
    :return: A nested dictionary
    :rtype: dict
    """
    for chunk in accessor_string.split('.')[::-1]:
        data = {chunk: data}
    return data


def memoize(func):
    """Memoization decorator for a function taking one or more arguments."""
    class MemoDict(dict):
        """Class to store outputs for previous inputs. If not previously input
        into func, gets added to the dict."""
        def __getitem__(self, *key):
            return dict.__getitem__(self, key)

        def __missing__(self, key):
            ret = self[key] = func(*key)
            return ret
    return MemoDict().__getitem__


def prevent_overwrite(write_path):
    """Prevents overwrite of existing output files by appending "_new" when
    needed

    :param write_path: potential write path
    :type write_path: string
    :return: new write path
    :rtype: str
    """
    while path.exists(write_path):
        split = write_path.split('.')
        # Make sure that files without an extension are still valid (otherwise,
        # split would create a list of one string which would give an index
        # error when sp[-2] is called)
        if len(split) > 1:
            split[-2] += '_new'
            write_path = '.'.join(split)
        else:
            write_path += '_new'
    return write_path


def approximate_matches(list1, list2, epsilon=0.01):
    """Takes two list of tuples and searches for matches of tuples first value
    within the supplied epsilon. Emits tuples with the tuples second values
    where found. if a value in one dist does not match the other list, it is
    emitted alone.

    :param list1: first list of tuples
    :type list1: list
    :param list2: second list of tuples
    :type list2: list
    :param epsilon: maximum difference in
    :type epsilon: float
    :return: second values of tuples
    :rtype: generator
    """
    list1.sort()
    list2.sort()
    list1_index = 0
    list2_index = 0

    while list1_index < len(list1) or list2_index < len(list2):
        if list1_index == len(list1):
            yield (0, list2[list2_index][1])
            list2_index += 1
            continue
        if list2_index == len(list2):
            yield (list1[list1_index][1], 0)
            list1_index += 1
            continue

        list1_element = list1[list1_index][0]
        list2_element = list2[list2_index][0]

        difference = abs(list1_element - list2_element)

        if difference < epsilon:
            yield (list1[list1_index][1], list2[list2_index][1])
            list1_index += 1
            list2_index += 1
        elif list1_element < list2_element:
            yield (list1[list1_index][1], 0)
            list1_index += 1
        elif list2_element < list1_element:
            yield (0, list2[list2_index][1])
            list2_index += 1
        else:
            raise AssertionError('Unexpected else taken')


def dict_merge(finaldict, sourcedict):
    """Merges two dictionaries using sets to avoid duplication of values"""
    for key, val in sourcedict.items():
        if (key in finaldict) and isinstance(finaldict[key], list):
            finaldict[key] = set(finaldict[key])
        if isinstance(val, list):
            if key in finaldict:
                finaldict[key].update(val)
            else:
                finaldict[key] = set(val)
        elif isinstance(val, str):
            if key in finaldict:
                finaldict[key].update(val)
            else:
                finaldict[key] = set(val)
                finaldict[key].update(val)
        elif isinstance(val, float):
            if key not in finaldict:
                finaldict[key] = val
        elif isinstance(val, dict):
            if key not in finaldict:
                finaldict[key] = {}
            dict_merge(finaldict[key], val)


def rxn2hash(reactants, products, return_text=False):
    """Hashes reactant and product lists"""
    # Get text reaction to be hashed
    def to_str(half_rxn):
        return ['(%s) %s' % (x[0], x[1]) if (len(x) == 2
                                             and not isinstance(x, str))
                else '(1) %s' % x for x in sorted(half_rxn)]

    text_rxn = ' + '.join(to_str(reactants)) + ' => ' + \
               ' + '.join(to_str(products))
    # Hash text reaction
    rhash = hashlib.sha256(text_rxn.encode()).hexdigest()
    if return_text:
        return rhash, text_rxn
    else:
        return rhash


def _calculate_rxn_hash(db, reactants, products):
    """Calculates a unique reaction hash using inchikeys. First block is
    connectivity only, second block is stereo only"""

    def __get_blocks(tups):
        first_block, second_block = [], []
        for x in tups:
            comp = db.compounds.find_one({"_id": x.c_id})
            if comp and comp["Inchikey"]:
                split_inchikey = comp["Inchikey"].split('-')
                if len(split_inchikey) > 1:
                    first_block.append("%s,%s" % (x.stoich, split_inchikey[0]))
                    second_block.append("%s,%s" % (x.stoich,
                                                   split_inchikey[1]))
            else:
                print("No Inchikey for %s" % x.c_id)
        return "+".join(first_block), "+".join(second_block)

    reactants.sort()
    products.sort()
    r_1, r_2 = __get_blocks(reactants)
    p_1, p_2 = __get_blocks(products)
    first_block = r_1 + '<==>' + p_1
    second_block = r_2 + '<==>' + p_2
    return hashlib.sha256(first_block.encode()).hexdigest() + "-" + \
        hashlib.md5(second_block.encode()).hexdigest()


def parse_text_rxn(rxn, rp_del, cp_del, translation_dict=None):
    """Makes a list of product and reactant StoichTuples"""

    def parse_half(half_rxn, t_d):
        if translation_dict:
            return [StoichTuple(1, t_d[x.strip()]) if len(x.split()) == 1
                    else StoichTuple(int(x.split()[0].strip('()')),
                                     t_d[x.split()[1].strip()])
                    for x in half_rxn.split(cp_del)]
        else:
            return [StoichTuple(1, x.strip()) if len(x.split()) == 1
                    else StoichTuple(int(x.split()[0].strip('()')),
                                     x.split()[1].strip())
                    for x in half_rxn.split(cp_del)]

    return [parse_half(x, translation_dict) for x in rxn.split(rp_del)]


_REACTIONS = None


def neutralise_charges(mol, reactions=None):
    """Neutralize all charges in compound (mol).

    :param mol: compound to neutralize
    :type mol: Mol object
    :param reactions: rules for neutralizing specific sites in the molecule
    :type reactions: tuple
    """
    def _initialise_neutralisation_reactions():
        patts = (
            # Imidazoles
            ('[n+;H]', 'n'),
            # Amines
            ('[N+;!H0]', 'N'),
            # Carboxylic acids and alcohols
            ('[$([O-]);!$([O-][#7])]', 'O'),
            # Thiols
            ('[S-;X1]', 'S'),
            # Sulfonamides
            ('[$([N-;X2]S(=O)=O)]', 'N'),
            # Enamines
            ('[$([N-;X2][C,N]=C)]', 'N'),
            # Tetrazoles
            ('[n-]', '[nH]'),
            # Sulfoxides
            ('[$([S-]=O)]', 'S'),
            # Amides
            ('[$([N-]C=O)]', 'N'),

        )
        return [(AllChem.MolFromSmarts(x), AllChem.MolFromSmiles(y, False))
                for x, y in patts]

    global _REACTIONS  # pylint: disable=global-statement
    if reactions is None:
        if _REACTIONS is None:
            _REACTIONS = _initialise_neutralisation_reactions()
        reactions = _REACTIONS
    for (reactant, product) in reactions:
        while mol.HasSubstructMatch(reactant):
            rms = AllChem.ReplaceSubstructs(mol, reactant, product)
            mol = rms[0]
    return mol


def score_compounds(db, compounds, model_id, parent_frac=0.75,
                    reaction_frac=0.25):
    """This function validates compounds against a metabolic model, returning
    only the compounds which pass.

    Parameters
    ----------
    db : Mongo DB
        Should contain a "models" collection with compound and reaction IDs
        listed.
    compounds : list
        Each element is a dict describing that compound. Should have an '_id'
        field.
    model_id : str
        KEGG organism code (e.g. 'hsa').
    parent_frac : float, optional (default: 0.75)
        Weighting for compounds derived from compounds in the provided model.
    reaction_frac : float, optional (default: 0.25)
        Weighting for compounds derived from known compounds not in the model.

    Returns
    -------
    compounds : list
        Modified version of input compounds list, where each compound now has
        a 'Likelihood_score' key and value between 0 and 1.
    """
    if not model_id:
        return compounds
    model = db.models.find_one({"_id": model_id})
    if not model:
        return compounds
    parents = set(model["Compounds"])

    for comp in compounds:
        try:
            if set(comp['DB_links']['KEGG']) & parents:
                comp['Likelihood_score'] = parent_frac + reaction_frac
                continue
        except KeyError:
            pass  # no worries if no KEGG id for this comp, just continue on

        if comp['Generation'] == 0:
            comp['Likelihood_score'] = reaction_frac
            continue

        comp['Likelihood_score'] = 0.0
        for source in comp['Sources']:
            likelihood_score = reaction_frac

            try:
                for s_comp in source['Compounds']:
                    if 'DB_links' in s_comp and 'KEGG' in s_comp['DB_links']:
                        if set(s_comp['DB_links']['KEGG']) & parents:
                            likelihood_score += parent_frac
            except KeyError:
                s_comp = source['Compound']  # needed for legacy MINEs
                if 'DB_links' in s_comp and 'KEGG' in s_comp['DB_links']:
                    if set(s_comp['DB_links']['KEGG']) & parents:
                        likelihood_score += parent_frac

            if likelihood_score > comp['Likelihood_score']:
                comp['Likelihood_score'] = likelihood_score

    return compounds


def get_smiles_from_mol_string(mol_string):
    """Convert a molfile in string format to a SMILES string."""
    mol = AllChem.MolFromMolBlock(mol_string)
    smiles = AllChem.MolToSmiles(mol)

    return smiles



"""This script reads a set of unknowns in a CSV, MGF, mzXML, or excel file and
queries a mongo database for potential matches. The matches are written back
to a CSV, HTML doc, or printed to STDOUT.

USAGE - metabolomics.py [options] run_name file_of_Unknowns
"""

import math
import pickle
import re
import sys
import os
import time
import xml.etree.ElementTree as ET
from optparse import OptionParser
from ast import literal_eval

import pandas as pd
import numpy as np

import minedatabase
from minedatabase.databases import establish_db_client
from minedatabase.utils import score_compounds


MINEDB_DIR = os.path.dirname(minedatabase.__file__)


class MetabolomicsDataset:
    """A class containing all the information for a metabolomics data set"""
    # pylint: disable=redefined-outer-name
    def __init__(self, name, options):
        self.name = name
        self.options = options  # Object containing all options from optparser

        dtype = np.dtype("U20, f8, f8")
        pos_fp = os.path.join(MINEDB_DIR,
                              'data/adducts/Positive Adducts full.txt')
        neg_fp = os.path.join(MINEDB_DIR,
                              'data/adducts/Negative Adducts full.txt')
        all_pos_adducts = np.loadtxt(pos_fp, dtype=dtype)
        all_neg_adducts = np.loadtxt(neg_fp, dtype=dtype)

        if hasattr(options, 'adducts') and options.adducts:
            pos_adducts = filter(lambda x: x[0] in options.adducts,
                                 all_pos_adducts)
            self.pos_adducts = np.array(list(pos_adducts), dtype=dtype)
            neg_adducts = filter(lambda x: x[0] in options.adducts,
                                 all_neg_adducts)
            self.neg_adducts = np.array(list(neg_adducts), dtype=dtype)
        else:
            if hasattr(options, 'positive_adduct_file') \
                and options.positive_adduct_file:
                self.pos_adducts = np.loadtxt(options.positive_adduct_file,
                                              dtype=dtype)
            else:
                self.pos_adducts = all_pos_adducts

            if hasattr(options, 'negative_adduct_file') \
                and options.negative_adduct_file:
                self.neg_adducts = np.loadtxt(options.negative_adduct_file,
                                              dtype=dtype)
            else:
                self.neg_adducts = all_neg_adducts

        if hasattr(options, 'kovats') and options.kovats:
            self.min_kovats = options.kovats[0]
            self.max_kovats = options.kovats[1]

        if hasattr(options, 'logp') and options.logp:
            self.min_logp = options.logp[0]
            self.max_logp = options.logp[1]

        self.hit_projection = {'Formula': 1, 'MINE_id': 1, 'logP': 1,
                               'minKovatsRI': 1, 'maxKovatsRI': 1,
                               'NP_likeness': 1, 'Names': 1, 'SMILES': 1,
                               'Inchikey': 1, 'Generation': 1,
                               'Pos_CFM_spectra': 1, 'Neg_CFM_spectra': 1,
                               "Sources": 1}
        self.known_peaks = []  # contains Peak objects for knowns
        self.unk_peaks = []  # contains Peak objects for unknowns
        self.clusters = []  # tuples of formula and list of matching peaks
        self.known_set = set()  # contains InChI key of all known compounds
        self.native_set = set()  # contains _ids of compounds in model set
        self.total_formulas = 0
        self.total_hits = 0
        self.matched_peaks = 0

    def __str__(self):
        return self.name

    def find_db_hits(self, peak, db, adducts):
        """This function searches the database for matches of a peak given
        adducts and updates a peak and dataset file with that information."""
        # find nominal mass for a given m/z for each adducts and the max and
        # min values for db
        potential_masses = (peak.mz - adducts['f2']) / adducts['f1']
        if self.options.ppm:
            precision = (self.options.tolerance / 100000.) * potential_masses
        else:
            precision = self.options.tolerance * 0.001
        upper_bounds = potential_masses + precision
        lower_bounds = potential_masses - precision

        # search database for hits in the each adducts mass range that have no
        # innate charge.
        for i, adduct in enumerate(adducts):
            # build the query by adding the optional terms
            query_terms = [{"Mass": {"$gte": float(lower_bounds[i])}},
                           {"Mass": {"$lte": float(upper_bounds[i])}},
                           {'Charge': 0}]
            if hasattr(self, 'min_logp'):
                query_terms += [{"logP": {"$gte": self.min_logp}},
                                {"logP": {"$lte": self.max_logp}}]
            if hasattr(self, 'min_kovats'):
                query_terms += [{"maxKovatsRI": {"$gte": self.min_kovats}},
                                {"minKovatsRI": {"$lte": self.max_kovats}}]
            if adduct['f0'] == '[M]+':
                query_terms[2] = {'Charge': 1}
            for compound in db.compounds.find({"$and": query_terms},
                                              self.hit_projection):
                # Filters out halogens if the flag is enabled by moving to the
                # next compound before the current compound is counted or
                # stored.
                if not self.options.halogens:
                    if re.search('F[^e]|Cl|Br', compound['Formula']):
                        continue

                # update the total hits for the peak and make a note if the
                # compound is in the native_set
                peak.total_hits += 1
                if compound['_id'] in self.native_set:
                    peak.native_hit = True
                    compound['native_hit'] = True
                if compound['Generation'] < peak.min_steps:
                    peak.min_steps = compound['Generation']
                peak.formulas.add(compound['Formula'])
                compound['adduct'] = adduct['f0']
                compound['peak_name'] = peak.name
                peak.isomers.append(compound)

    def annotate_peaks(self, db):
        """This function iterates the through the unknown peaks in the dataset
        searches the database for compounds that match a peak m/z given the
        adducts permitted. Statistics on the annotated data set are printed"""
        for i, peak in enumerate(self.unk_peaks):

            positive = peak.charge == '+' or peak.charge == 'Positive' \
                or (peak.charge and isinstance(peak.charge, bool))
            negative = peak.charge == '-' or peak.charge == 'Negative' \
                or (not peak.charge and isinstance(peak.charge, bool))

            if positive:
                self.find_db_hits(peak, db, self.pos_adducts)
            elif negative:
                self.find_db_hits(peak, db, self.neg_adducts)
            else:
                raise ValueError("Invalid compound charge specification. "
                                 "Please use \"+\" or \"Positive\" for "
                                 "positive ions and \"-\" or \"Negative\" for "
                                 f"negative ions. (charge = {peak.charge})")

            if peak.total_hits > 0:
                self.matched_peaks += 1
                self.total_hits += peak.total_hits
                self.total_formulas += len(peak.formulas)
            if self.options.verbose:
                pct_done = int(float(i) / float(len(self.unk_peaks)) * 100)
                print("%s percent of peaks processed" % pct_done)

        if self.options.verbose:
            print("Proposed matches for %s of %s peaks"
                  % (self.matched_peaks, len(self.unk_peaks)))

            try:
                print("Average hits per peak: %s"
                      % (float(self.total_hits) / float(self.matched_peaks)))
                print("Average formulas per peak: %s"
                      % (float(self.total_formulas)
                         / float(self.matched_peaks)))
            except ZeroDivisionError:
                pass


def dot_product(x, y, epsilon=0.01):
    """Calculate the dot_product of two spectra"""
    z = 0
    n_v1 = 0
    n_v2 = 0

    for int1, int2 in approximate_matches(x, y, epsilon):
        z += int1 * int2
        n_v1 += int1 * int1
        n_v2 += int2 * int2

    return z / (math.sqrt(n_v1) * math.sqrt(n_v2))


def jaccard(x, y, epsilon=0.01):
    """Calculate the Jaccard Index of two spectra"""
    intersect = 0

    for val1, val2 in approximate_matches(x, y, epsilon):
        if val1 and val2:
            intersect += 1

    return intersect / float((len(x) + len(y) - intersect))


def approximate_matches(list1, list2, epsilon=0.01):
    """
    Takes two list of tuples and searches for matches of tuples first value
    within the supplied epsilon. Emits tuples with the tuples second values
    where found. if a value in one dist does not match the other list, it is
    emitted alone.
    :param list1: first list of tuples
    :type list1: list
    :param list2: second list of tuples
    :type list2: list
    :param epsilon: maximum difference in
    :type epsilon: float
    :return: second values of tuples
    :rtype: generator
    """
    list1.sort()
    list2.sort()
    list1_index = 0
    list2_index = 0

    while list1_index < len(list1) or list2_index < len(list2):
        if list1_index == len(list1):
            yield (0, list2[list2_index][1])
            list2_index += 1
            continue
        if list2_index == len(list2):
            yield (list1[list1_index][1], 0)
            list1_index += 1
            continue

        list1_element = list1[list1_index][0]
        list2_element = list2[list2_index][0]

        difference = abs(list1_element - list2_element)

        if difference < epsilon:
            yield (list1[list1_index][1], list2[list2_index][1])
            list1_index += 1
            list2_index += 1
        elif list1_element < list2_element:
            yield (list1[list1_index][1], 0)
            list1_index += 1
        elif list2_element < list1_element:
            yield (0, list2[list2_index][1])
            list2_index += 1
        else:
            raise AssertionError('Unexpected else taken')


class Peak:
    """A class holding information about an unknown peak"""
    def __init__(self, name, r_time, mz, charge, inchi_key, ms2=[]):
        self.name = name
        self.r_time = float(r_time)  # retention time
        self.mz = float(mz)  # mass to charge ratio
        self.charge = charge  # polarity of charge
        self.inchi_key = inchi_key  # Inchikey, if known
        self.ms2peaks = ms2
        self.isomers = []
        self.formulas = set()
        self.total_hits = 0
        self.native_hit = False
        self.min_steps = 99

    def __str__(self):
        return self.name

    def score_isomers(self, metric=dot_product, energy_level=20,
                      tolerance=0.005):
        """
        Calculates the cosign similarity score between the provided ms2 peak
        list and pre-calculated CFM-spectra and sorts the isomer list
        according to this metric.
        :param metric: The scoring metric to use for the spectra. Function
         must accept 2 lists of (mz,intensity) tuples and return a score.
         Defaults to dot_product.
        :type energy_level: function
        :param energy_level: The Fragmentation energy level to use. May be 10,
         20 or 40. Defaults to 20
        :type energy_level: int
        :param tolerance: The precision to use for matching m/z in mDa
        :type energy_level: float
        :return:
        :rtype:
        """
        if not self.ms2peaks:
            raise ValueError('The ms2 peak list is empty')
        if self.charge:
            spec_key = "Pos_CFM_spectra"
        else:
            spec_key = "Neg_CFM_spectra"

        for i, hit in enumerate(self.isomers):
            if spec_key in hit:
                hit_spec = hit[spec_key]['%s V' % energy_level]
                score = metric(self.ms2peaks, hit_spec, epsilon=tolerance)
                rounded_score = round(score * 1000)
                self.isomers[i]['Spectral_score'] = rounded_score
                del hit[spec_key]
            else:
                self.isomers[i]['Spectral_score'] = None
        self.isomers.sort(key=lambda x: x['Spectral_score'], reverse=True)


def get_modelseed_comps(kb_db, model_ids):
    """Get MongoIDs from KBase database for compounds in model(s)."""
    comp_ids, _ids = set(), set()
    for model_id in model_ids:
        comp_ids = kb_db.models.find_one({'_id': model_id},
                                         {'Compounds': 1})['Compounds']
        for comp_id in comp_ids:
            comp_ids.add(comp_id)
    for comp_id in comp_ids:
        comp_id = int(comp_id.split('.')[1])
        for comp in kb_db.compounds.find({'Model_SEED': comp_id}, {'_id': 1}):
            _ids.add(comp['_id'])
    return _ids


def get_KEGG_comps(db, kegg_db, model_ids):
    """Get KEGG IDs from KEGG MINE database for compounds in model(s)."""
    kegg_ids, _ids = set(), set()
    for model_id in model_ids:
        comp_ids = kegg_db.models.find_one({'_id': model_id},
                                           {'Compounds': 1})['Compounds']
        for comp_id in comp_ids:
            kegg_ids.add(comp_id)
    for kegg_id in kegg_ids:
        for comp in db.compounds.find({'DB_links.KEGG': kegg_id}, {'_id': 1}):
            _ids.add(comp['_id'])
    return _ids


def sort_nplike(dic_list):
    """Sort in descending order of natural product likeness."""
    return sorted(dic_list, key=lambda x: float(x['NP_likeness']),
                  reverse=True)


def read_adduct_names(filepath):
    """Read adduct names from text file at specified path into a list.

    Parameters
    ----------
    filepath : str
        Path to adduct text file.

    Returns
    -------
    adducts : list
        Names of adducts in text file.
    """

    with open(filepath) as infile:
        adducts = [line.split(' \t')[0] for line in infile
                   if not line[0] == '#']

    return adducts


def read_mgf(input_string, charge):
    """Parse mgf metabolomics data file."""
    peaks = []
    ms2 = []
    for line in input_string.split('\n'):
        sl = line.strip(' \r\n').split('=')
        if sl[0] == "PEPMASS":
            mass = sl[1]
        elif sl[0] == "TITLE":
            name = sl[1]
        elif sl[0] == "RTINSECONDS":
            r_time = sl[1]
        elif sl[0] == "END IONS":
            peaks.append(Peak(name, r_time, mass, charge, "False", ms2=ms2))
            ms2 = []
        else:
            try:
                mz, i = sl[0].split('\t')
                ms2.append((float(mz), float(i)))
            except ValueError:
                continue
    return peaks


def read_msp(input_string, charge):
    """Parse msp metabolomics data file."""
    peaks = []
    for spec in input_string.strip().split('\n\n'):
        ms2 = []
        inchikey = "False"
        r_time = 0
        name = 'N/A'
        for line in spec.split('\n'):
            sl = line.split(': ')
            sl[0] = sl[0].replace(' ', '').replace('/', '').upper()
            if sl[0] == "PRECURSORMZ":
                mass = sl[1]
            elif sl[0] == "NAME":
                name = sl[1]
            # elif sl[0] == "RETENTIONTIME":
                # r_time = sl[1]
            # elif sl[0] == "IONMODE":
                # charge = sl[1].capitalize()
            elif sl[0] == "INCHIKEY":
                inchikey = sl[1]
            elif line and line[0].isdigit():
                try:
                    row = re.split('[\t ]', line)
                    ms2.append((float(row[0]), float(row[1])))
                except ValueError:
                    continue
        peaks.append(Peak(name, r_time, mass, charge, inchikey, ms2=ms2))
    return peaks


def read_mzxml(input_string, charge):
    """Parse mzXML metabolomics data file."""
    peaks = []
    root = ET.fromstring(input_string)
    prefix = root.tag.strip('mzXML')

    for scan in root.findall('.//%sscan' % prefix):
        # somewhat counter intuitively we will get the peak info from the
        # second fragments precursor info.
        if scan.attrib['msLevel'] == '2':
            precursor = scan.find('./%sprecursorMz' % prefix)
            mz = precursor.text
            r_time = scan.attrib['retentionTime'][2:-1]
            name = "%s @ %s" % (mz, r_time)
            charge = scan.attrib['polarity']
            peaks.append(Peak(name, r_time, mz, charge, "False"))

    return peaks


class Struct:
    """convert key-value pairs into object-attribute pairs."""
    def __init__(self, **entries):
        self.__dict__.update(entries)


def ms_adduct_search(db, keggdb, text, text_type, ms_params):
    """Search for compound-adducts matching precursor mass.

    Parameters
    ----------
    db : Mongo DB
        Contains compound documents to search.
    keggdb : Mongo DB
        Contains models with associated compound documents.
    text : str
        Text as in metabolomics datafile for specific peak.
    text_type : str
        Type of metabolomics datafile (mgf, mzXML, and msp are supported). If
        text, assumes m/z values are separated by newlines (and set text_type
        to 'form').
    ms_params : dict
        Specifies search settings, using the following key-value pairs:
        ------------------------
        Required Key-Value Pairs
        ------------------------
        'tolerance': float specifying tolerance for m/z, in mDa by default.
            Can specify in ppm if 'ppm' key's value is set to True.
        'charge': bool (1 for positive, 0 for negative).
        ------------------------
        Optional Key-Value Pairs
        ------------------------
        'adducts': list of adducts to use. If not specified, uses all adducts.
        'models': List of model _ids. If supplied, score compounds higher if
            present in model. ['eco'] by default (E. coli).
        'ppm': bool specifying whether 'tolerance' is in mDa or ppm. Default
            value for ppm is False (so tolerance is in mDa by default).
        'kovats': length 2 tuple specifying min and max kovats retention index
            to filter compounds (e.g. (500, 1000)).
        'logp': length 2 tuple specifying min and max logp to filter compounds
            (e.g. (-1, 2)).
        'halogens': bool specifying whether to filter out compounds containing
            F, Cl, or Br. Filtered out if set to True. False by default.

    Returns
    -------
    ms_adduct_output : list
        Compound JSON documents matching ms adduct query.
    """
    print("<MS Adduct Search: TextType=%s, Text=%s, Parameters=%s>"
          % (text_type, text, ms_params))
    name = text_type + time.strftime("_%d-%m-%Y_%H:%M:%S", time.localtime())

    if isinstance(ms_params, dict):
        ms_params = Struct(**ms_params)

    dataset = MetabolomicsDataset(name, ms_params)
    ms_adduct_output = []

    if text_type == 'form':
        for mz in text.split('\n'):
            dataset.unk_peaks.append(Peak(mz, 0, float(mz), ms_params.charge,
                                          "False"))
    elif text_type == 'mgf':
        dataset.unk_peaks = read_mgf(text, ms_params.charge)
    elif text_type == 'mzXML' or text_type == 'mzxml':
        dataset.unk_peaks = read_mzxml(text, ms_params.charge)
    elif text_type == 'msp':
        dataset.unk_peaks = read_msp(text, ms_params.charge)
    else:
        raise IOError('%s files not supported' % text_type)

    if not ms_params.models:
        ms_params.models = ['eco']

    dataset.native_set = get_KEGG_comps(db, keggdb, ms_params.models)
    dataset.annotate_peaks(db)

    for peak in dataset.unk_peaks:
        for hit in peak.isomers:
            if 'CFM_spectra' in hit:
                del hit['CFM_spectra']
            ms_adduct_output.append(hit)

    ms_adduct_output = score_compounds(db, ms_adduct_output,
                                       ms_params.models[0],
                                       parent_frac=.75, reaction_frac=.25)

    return ms_adduct_output


def ms2_search(db, keggdb, text, text_type, ms_params):
    """Search for compounds matching MS2 spectra.

    Parameters
    ----------
    db : Mongo DB
        Contains compound documents to search.
    keggdb : Mongo DB
        Contains models with associated compound documents.
    text : str
        Text as in metabolomics datafile for specific peak.
    text_type : str
        Type of metabolomics datafile (mgf, mzXML, and msp are supported). If
        text, assumes m/z values are separated by newlines (and set text_type
        to 'form').
    ms_params : dict
        Specifies search settings, using the following key-value pairs:
        ------------------------
        Required Key-Value Pairs
        ------------------------
        'tolerance': float specifying tolerance for m/z, in mDa by default.
            Can specify in ppm if 'ppm' key's value is set to True.
        'charge': bool (1 for positive, 0 for negative).
        'energy_level': int specifying fragmentation energy level to use. May
            be 10, 20, or 40.
        'scoring_function': str describing which scoring function to use. Can
            be either 'jaccard' or 'dot product'.
        ------------------------
        Optional Key-Value Pairs
        ------------------------
        'adducts': list of adducts to use. If not specified, uses all adducts.
        'models': List of model _ids. If supplied, score compounds higher if
            present in model.
        'ppm': bool specifying whether 'tolerance' is in mDa or ppm. Default
            value for ppm is False (so tolerance is in mDa by default).
        'kovats': length 2 tuple specifying min and max kovats retention index
            to filter compounds (e.g. (500, 1000)).
        'logp': length 2 tuple specifying min and max logp to filter compounds
            (e.g. (-1, 2)).
        'halogens': bool specifying whether to filter out compounds containing
            F, Cl, or Br. Filtered out if set to True. False by default.

    Returns
    -------
    ms_adduct_output : list
        Compound JSON documents matching ms2 search query.
    """
    print("<MS Adduct Sea""rch: TextType=%s, Parameters=%s>"
          % (text_type, ms_params))
    name = text_type + time.strftime("_%d-%m-%Y_%H:%M:%S", time.localtime())

    if isinstance(ms_params, dict):
        ms_params = Struct(**ms_params)

    dataset = MetabolomicsDataset(name, ms_params)
    ms_adduct_output = []

    if text_type == 'form':
        split_form = [x.split() for x in text.strip().split('\n')]
        ms2_data = [(float(mz), float(i)) for mz, i in split_form[1:]]
        peak = Peak(split_form[0][0], 0, float(split_form[0][0]),
                    ms_params.charge, "False", ms2=ms2_data)
        dataset.unk_peaks.append(peak)
    elif text_type == 'mgf':
        dataset.unk_peaks = read_mgf(text, ms_params.charge)
    elif text_type == 'mzXML':
        dataset.unk_peaks = read_mzxml(text, ms_params.charge)
    elif text_type == 'msp':
        dataset.unk_peaks = read_msp(text, ms_params.charge)
    else:
        raise IOError('%s files not supported' % text_type)

    if not ms_params.models:
        ms_params.models = ['eco']

    dataset.native_set = get_KEGG_comps(db, keggdb, ms_params.models)
    dataset.annotate_peaks(db)

    for peak in dataset.unk_peaks:

        if ms_params.scoring_function == 'jaccard':
            if not ms_params.ppm:
                peak.score_isomers(metric=jaccard,
                                   energy_level=ms_params.energy_level,
                                   tolerance=float(ms_params.tolerance) / 1000)
            else:
                peak.score_isomers(metric=jaccard,
                                   energy_level=ms_params.energy_level)
        elif ms_params.scoring_function == 'dot product':
            if not ms_params.ppm:
                peak.score_isomers(metric=dot_product,
                                   energy_level=ms_params.energy_level,
                                   tolerance=float(ms_params.tolerance) / 1000)
            else:
                peak.score_isomers(metric=dot_product,
                                   energy_level=ms_params.energy_level)
        else:
            raise ValueError("ms_params['scoring_function'] must be either "
                             "'jaccard' or 'dot product'.")

        for hit in peak.isomers:
            ms_adduct_output.append(hit)

        if ms_params.models:
            ms_adduct_output = score_compounds(db, ms_adduct_output,
                                               ms_params.models[0],
                                               parent_frac=.75,
                                               reaction_frac=.25)

    return ms_adduct_output


def spectra_download(db, mongo_query=None, parent_filter=False, putative=True):
    """Download one or more spectra for compounds matching a given query.

    Parameters
    ----------
    db : Mongo DB
        Contains compound documents to search.
    mongo_query : str, optional (default: None)
        A valid Mongo query as a literal string. If None, all compound spectra
        are returned.
    parent_filter : str, optional (default: None)
        If set to a metabolic model's Mongo _id, only get spectra for compounds
        in or derived from that metabolic model.
    putative : bool, optional (default: True)
        If False, only find known compounds (i.e. in Generation 0). Otherwise,
        finds both known and predicted compounds.

    Returns
    -------
    spectral_library : str
        Text of all matching spectra, including headers and peak lists.
    """
    def print_peaklist(peaklist):
        text = ["Num Peaks: %s" % len(peaklist)]
        for x in peaklist:
            text.append("%s %s" % (x[0], x[1]))
        text.append("")
        return text

    spectral_library = []
    msp_projection = {'MINE_id': 1, 'Names': 1, 'Mass': 1, 'Generation': 1,
                      'Inchikey': 1, 'Formula': 1, 'SMILES': 1, 'Sources': 1,
                      'Pos_CFM_spectra': 1, 'Neg_CFM_spectra': 1}

    if mongo_query:
        query_dict = literal_eval(mongo_query)
    else:
        query_dict = {}

    if not putative:
        query_dict['Generation'] = 0

    if parent_filter:
        model = db.models.find_one({"_id": parent_filter})

        if not model:
            raise ValueError('Invalid Model specified')

        parents = model["Compound_ids"]
        query_dict['$or'] = [{'_id': {'$in': parents}},
                             {'Sources.Compound': {'$in': parents}}]

    results = db.compounds.find(query_dict, msp_projection)

    for compound in results:

        # add header
        header = []
        if "Names" in compound and len(compound['Names']) > 0:
            header.append("Name: %s" % compound['Names'][0])
            for alt in compound['Names'][1:]:
                header.append("Synonym: %s" % alt)
        else:
            header.append("Name: MINE Compound %s" % compound['MINE_id'])

        for k, v in compound.items():
            if k not in {"Names", "Pos_CFM_spectra", "Neg_CFM_spectra"}:
                header.append("%s: %s" % (k, v))

        header.append("Instrument: CFM-ID")

        # add peak lists
        if 'Pos_CFM_spectra' in compound:
            for energy, spec in compound['Pos_CFM_spectra'].items():
                # ??? not sure what James meant with this conditional:
                # he had spec_type as a argument to this function
                # if not spec_type or [True, int(energy[:2])] in spec_type:
                spectral_library += header
                spectral_library += ["Ionization: Positive",
                                     "Energy: %s" % energy]
                spectral_library += print_peaklist(spec)

        if 'Neg_CFM_spectra' in compound:
            for energy, spec in compound['Neg_CFM_spectra'].items():
                # ??? not sure what James meant with this conditional:
                # if not spec_type or [False, int(energy[:2])] in spec_type:
                spectral_library += header
                spectral_library += ["Ionization Mode: Negative",
                                     "Energy: %s" % energy]
                spectral_library += print_peaklist(spec)

    spectral_library = "\n".join(spectral_library)

    return spectral_library


if __name__ == '__main__':
    # pylint: disable=invalid-name

    sys.exit()
    t_start = time.time()
    # This block handles user flags and arguments. For more information see the
    # optparse API documentation

    usage = "usage: %prog [options] run_name"
    parser = OptionParser(usage)
    parser.add_option("-c", "--charge", dest="unknowns_charge",
                      default='Positive', help='Charge for unknowns. Options '
                      'are "Positive" or "Negative" (case-sensitive).')
    parser.add_option("-p", "--positive_adducts", dest="positive_adduct_file",
                      default="Batch Adduct Query/Positive Adducts.txt",
                      help="The path to the desired positive adducts file")
    parser.add_option("-n", "--negative_adducts", dest="negative_adduct_file",
                      default="Batch Adduct Query/Negative Adducts.txt",
                      help="The path to the desired negative adducts file")
    parser.add_option("-d", "--database", dest="database", default="1GenKEGG",
                      help="The name of the database to search")
    parser.add_option("-m", "--modelSEED", dest="modelSEED", default="null",
                      help="The model SEED id of the organism for the data "
                      "set")
    parser.add_option("-k", "--known", dest="known_file", default="null",
                      help="The path to a file containing known or targeted "
                      "peaks")
    parser.add_option("-t", "--tolerance", dest="tolerance", type="float",
                      default=2, help="The m/z tolerance(precision) for peak "
                      "searching")
    parser.add_option("--ppm", dest="ppm", action="store_true", default=False,
                      help="Tolerance is in Parts Per Million")
    parser.add_option("-x", "--halogens", dest="halogens", action="store_true",
                      default=False, help="include compounds containing F, "
                      "Cl, and Br")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                      default=False, help="include compounds containing F, "
                      "Cl, and Br")

    (options, args) = parser.parse_args()
    # Ensure that the right number of arguments are passed to the script.
    if len(args) < 2:
        sys.exit("Missing required arguments Usage: [options] run_name, "
                 "path_to_unk_peaks")
    if len(args) > 2:
        sys.exit("Too many arguments submitted. Usage: [options] run_name, "
                 "path_to_unk_peaks")

    data = MetabolomicsDataset(args[0], options)
    unknowns_file = args[1]

    # if the user specifies known peaks import them for use in clustering and
    # ordering of isomers
    if options.known_file != 'null':
        print("Loading known peaks")
        data.known_peaks = list(pd.read_csv(options.known_file))
        for peak in data.known_peaks:
            data.known_set.add(peak.known)

    # detect unknown compounds file type and parse accordingly
    if ('.txt' or '.csv') in unknowns_file:
        data.unk_peaks = list(pd.read_csv(open(unknowns_file).read()))
    elif '.mgf' in unknowns_file:
        data.unk_peaks = read_mgf(open(unknowns_file).read(), options.charge)
    elif '.mzXML' in unknowns_file:
        data.unk_peaks = read_mzxml(open(unknowns_file).read(), options.charge)
    else:
        sys.exit("Unknown file type not recognised. Please use .mgf, .xlsx, "
                 ".txt or .csv file")

    client = establish_db_client()
    db = client[options.database]
    if not db.compounds.count():
        sys.exit('No compounds in supplied database')

    if options.modelSEED != 'null':
        print("Loading native compounds")
        kbase_db = client['KBase']
        data.native_set = get_modelseed_comps(kbase_db, [options.modelSEED])

    data.annotate_peaks(db)
    with open(data.name + '.pkl') as outfile:
        pickle.dump(data, outfile)

    t_end = time.time()
    print("BatchAdductQuery.py completed in %s seconds" % (t_end - t_start))



"""Tests for queries.py using pytest."""
# pylint: disable=redefined-outer-name

import os
from collections import OrderedDict
import minedatabase.utils as utils

DATA_DIR = os.path.dirname(__file__) + '/data/'


def test_file_to_dict_list():
    """
    GIVEN compound files of different types (.tsv, .csv, and .json)
    WHEN the file contents are converted to a list of compound dicts
    THEN check that the list of compound dicts is produced as expected
    """
    res_7 = OrderedDict(
        [('id', 'cpd01211'), ('abbreviation', 'tcynt'),
         ('name', 'Thiocyanate'), ('formula', 'CNS'), ('mass', '58'),
         ('source', 'ModelSEED'), ('structure', 'InChI=1S/CHNS/c2-1-3/h3H'),
         ('charge', '-1'), ('is_core', '1'), ('is_obsolete', '0'),
         ('linked_compound', 'null'), ('is_cofactor', '0'), ('deltag', '22.2'),
         ('deltagerr', '5.68687'), ('pka', '3:0.5'), ('pkb', ''),
         ('abstract_compound', 'null'), ('comprised_of', 'null'),
         ('aliases', 'null')])

    filenames = ['test_compounds.tsv', 'test_compounds.csv',
                 'test_compounds.json']

    for file in filenames:
        res = utils.file_to_dict_list(DATA_DIR + file)
        assert len(res) == 15
        assert res[7] == res_7



"""Queries.py: Contains functions which power the API queries"""
import re
from ast import literal_eval
from minedatabase.utils import score_compounds

from rdkit.Chem import AllChem

DEFAULT_PROJECTION = {'SMILES': 1, 'Formula': 1, 'MINE_id': 1, 'Names': 1,
                      'Inchikey': 1, 'Mass': 1, 'Sources': 1, 'Generation': 1,
                      'NP_likeness': 1, 'DB_links': 1}


def quick_search(db, query, search_projection=DEFAULT_PROJECTION.copy()):
    """This function takes user provided compound identifiers and attempts to
     find a related database ID

    :param db: DB to search
    :type db: A Mongo Database
    :param query: A MINE id, KEGG code, ModelSEED id, Inchikey or Name
    :type query: str
    :param search_projection: The fields which should be returned
    :type search_projection: str
    :return: Query results
    :rtype: list
    """

    # Determine what kind of query was input (e.g. KEGG code, MINE id, etc.)
    # If it can't be determined to be an id or a key, assume it is the name
    # of a compound.
    if re.match(r"C\w{40}", query):
        query_field = '_id'
    elif re.match(r"C\d{5}", query):
        query_field = 'DB_links.KEGG'
    elif re.match(r'cpd\d{5}', query):
        query_field = 'DB_links.Model_SEED'
    elif re.search(r"[A-Z]{14}-[A-Z]{10}-[A-Z]", query):
        query_field = 'Inchikey'
        query = query.split("=", 1)[-1]
    elif query.isdigit():
        query_field = "MINE_id"
        query = int(query)
    else:
        query_field = 'Names'

    if query_field == 'Names':
        # Return results for all compounds with specified name
        # Make sure that cofactors are not included
        results = db.compounds.find({"Names": {'$regex': '^' + query + '$',
                                               '$options': 'i'}},
                                    search_projection)
        results = [x for x in results if x['_id'][0] == "C"]
        if not results:
            cursor = db.compounds.find({"$text": {"$search": query}},
                                       {"score": {"$meta": "textScore"},
                                        'Formula': 1, 'MINE_id': 1, 'Names': 1,
                                        'Inchikey': 1, 'SMILES': 1, 'Mass': 1})
            top_x = cursor.sort([("score", {"$meta": "textScore"})]).limit(500)
            results.extend(x for x in top_x if x['_id'][0] == "C")
    else:
        # If query isn't a compound name, then return compounds with
        # specified query
        results = [x for x in db.compounds.find({query_field: query},
                                                search_projection).limit(500)
                   if x['_id'][0] == "C"]

    return results


def advanced_search(db, mongo_query,
                    search_projection=DEFAULT_PROJECTION.copy()):
    """Returns compounds in the indicated database which match the provided
    mongo query

    :param db: DB to search
    :type db: A Mongo Database
    :param query: A valid mongo query
    :type query: str
    :param search_projection: The fields which should be returned
    :type search_projection: str
    :return: Query results
    :rtype: list
    """
    # We don't want users poking around here
    if db.name == 'admin' or not mongo_query:
        raise ValueError('Illegal query')
    # This transforms the string into a dictionary
    query_dict = literal_eval(mongo_query)
    return [x for x in db.compounds.find(query_dict, search_projection)]


def similarity_search(db, comp_structure, min_tc, limit, parent_filter=None,
                      model_db=None, fp_type='RDKit',
                      search_projection=DEFAULT_PROJECTION.copy()):
    """Returns compounds in the indicated database which have structural
     similarity to the provided compound

    :param db: DB to search
    :type db: A Mongo Database
    :param comp_structure: A molecule in Molfile or SMILES format
    :type comp_structure: str
    :param min_tc: Minimum Tanimoto score
    :type min_tc: float
    :param fp_type: Fingerprint type. Currently accepts MACCS or RDKit
    :type fp_type: str
    :param limit: The maximum number of compounds to return
    :type limit: int
    :param parent_filter: str
    :type parent_filter: str
    :param search_projection: The fields which should be returned
    :type search_projection: str
    :return: Query results
    :rtype: list
    """
    similarity_search_results = []
    fp_type = str(fp_type)
    # Create Mol object from Molfile (has newlines)
    if "\n" in comp_structure:
        mol = AllChem.MolFromMolBlock(str(comp_structure))
    # Create Mol object from SMILES string (does not have newlines)
    else:
        mol = AllChem.MolFromSmiles(str(comp_structure))
    if not mol:
        raise ValueError("Unable to parse comp_structure")

    # Based on fingerprint type specified by user, get the finger print as an
    # explicit bit vector (series of 1s and 0s). Then, return a set of all
    # indices where a bit is 1 in the bit vector.
    if fp_type == 'MACCS':
        query_fp = set(AllChem.GetMACCSKeysFingerprint(mol).GetOnBits())
    elif fp_type == 'RDKit':
        query_fp = set(AllChem.RDKFingerprint(mol).GetOnBits())
    else:
        raise ValueError("Invalid FP_type")

    len_fp = len(query_fp)
    # Return only id and fingerprint vector
    search_projection[fp_type] = 1

    # Filter compounds that meet tanimoto coefficient size requirements
    for x in db.compounds.find(
            {"$and": [{"len_" + fp_type: {"$gte": min_tc * len_fp}},
                      {"len_" + fp_type: {"$lte": len_fp / min_tc}}]},
            search_projection):
        # Put fingerprint in set for fast union (&) and intersection (|)
        # calculations
        test_fp = set(x[fp_type])
        # Calculate tanimoto coefficient
        tmc = len(query_fp & test_fp) / float(len(query_fp | test_fp))
        # If a sufficient tanimoto coefficient is calculated, append the
        # compound to the search results (until the limit is reached)
        if tmc >= min_tc:
            del x[fp_type]
            similarity_search_results.append(x)
            if len(similarity_search_results) == limit:
                break

    del search_projection[fp_type]

    if parent_filter and model_db:
        similarity_search_results = score_compounds(model_db,
                                                    similarity_search_results,
                                                    parent_filter)

    return similarity_search_results


def structure_search(db, comp_structure, stereo=True, parent_filter=None,
                     model_db=None,
                     search_projection=DEFAULT_PROJECTION.copy()):
    """Returns compounds in the indicated database which are exact matches to
    the provided structure

    :param db: DB to search
    :type db: A Mongo Database
    :param comp_structure: A molecule in Molfile or SMILES format
    :type comp_structure: str
    :param stereo: If true, uses stereochemistry in finding exact match
    :type stereo: bool
    :param search_projection: The fields which should be returned
    :type search_projection: str
    :return: Query results
    :rtype: list    """
    # Create Mol object from Molfile (has newlines)
    if "\n" in comp_structure:
        mol = AllChem.MolFromMolBlock(str(comp_structure))
    # Create Mol object from SMILES string (does not have newlines)
    else:
        mol = AllChem.MolFromSmiles(str(comp_structure))
    if not mol:
        raise ValueError("Unable to parse comp_structure")

    # Get InChI string from mol file (rdkit)
    inchi = AllChem.MolToInchi(mol)
    # Get InChI key from InChI string (rdkit)
    inchi_key = AllChem.InchiToInchiKey(inchi)
    # Sure, we could look for a matching SMILES but this is faster
    if stereo:
        results = quick_search(db, inchi_key, search_projection)
    else:
        results = [x for x in db.compounds.find(
            {"Inchikey": {'$regex': '^' + inchi_key.split('-')[0]}},
            search_projection)]

    if parent_filter and model_db:
        results = score_compounds(model_db, results, parent_filter)

    return results


def substructure_search(db, sub_structure, limit, parent_filter=None,
                        model_db=None,
                        search_projection=DEFAULT_PROJECTION.copy()):
    """Returns compounds in the indicated database which contain the provided
    structure

    :param db: DB to search
    :type db: A Mongo Database
    :param comp_structure: A molecule in Molfile or SMILES format
    :type comp_structure: str
    :param limit: The maximum number of compounds to return
    :type limit: int
    :param search_projection: The fields which should be returned
    :type search_projection: str
    :return: Query results
    :rtype: list
    """
    substructure_search_results = []
    # Create Mol object from Molfile (has newlines)
    if "\n" in sub_structure:
        mol = AllChem.MolFromMolBlock(str(sub_structure))
    # Create Mol object from SMILES string (does not have newlines)
    else:
        mol = AllChem.MolFromSmiles(str(sub_structure))
    if not mol:
        raise ValueError("Unable to parse comp_structure")

    # Based on fingerprint type specified by user, get the finger print as an
    # explicit bit vector (series of 1s and 0s). Then, return a set of all
    # indices where a bit is 1 in the bit vector.
    query_fp = list(AllChem.RDKFingerprint(mol).GetOnBits())
    for x in db.compounds.find({"RDKit": {"$all": query_fp}},
                               search_projection):
        # Get Mol object from SMILES string (rdkit)
        comp = AllChem.MolFromSmiles(x['SMILES'])
        # Use HasSubstructMatch (rdkit) to determine if compound has a
        # specified substructure. If so, append it to the results (until
        # limit).
        if comp and comp.HasSubstructMatch(mol):
            substructure_search_results.append(x)
            if len(substructure_search_results) == limit:
                break

    if parent_filter and model_db:
        substructure_search_results = \
            score_compounds(model_db, substructure_search_results,
                            parent_filter)
 
    return substructure_search_results


def model_search(db, query):
    """Returns models that match a given KEGG Org Code query (e.g. 'hsa').

    Parameters
    ----------
    db : Mongo DB
        DB to search. Should contain a 'models' collection. '_id' and 'Name'
        fields should be indexed for text search.
    query : str
        KEGG Org Code or Org Name of model(s) to search for (e.g. 'hsa' or
        'yeast'). Can provide multiple search terms by separating each term
        with a space.  TODO: change from space delimiter to something else

    Returns
    -------
    model_ids : str
        List of KEGG Org codes (_id) found in the DB matching search query.
    """
    cursor = db.models.find({"$text": {"$search": query}},
                            {"score": {"$meta": "textScore"}, "_id": 1})
    model_ids = [x["_id"] for x in cursor.sort([("score",
                                                 {"$meta": "textScore"})])]
    return model_ids


def get_ids(db, collection, query):
    """Returns ids for documents in database collection matching query.

    :param db: DB to search
    :type db: A Mongo Database
    :param collection: collection within DB to search
    :type collection: A Mongo Database Collection
    :param query: A valid Mongo query.
    :type query: str
    """

    if query:
        query = literal_eval(query)
    else:
        query = {}
    ids = [x['_id'] for x in db[collection].find(query)]

    return ids


def get_comps(db, id_list):
    """Returns compounds with associated IDs from a Mongo database.

    :param db: DB to search
    :type db: A Mongo Database
    :param id_list: IDs to get compounds for.
    :type id_list: list
    :return: Compound JSON documents
    :rtype: list
    """
    compounds = []
    for cpd_id in id_list:
        excluded_fields = {"len_FP2": 0, "FP2": 0, "len_FP4": 0, "FP4": 0}
        if isinstance(cpd_id, int):
            cpd = db.compounds.find_one({'MINE_id': cpd_id}, excluded_fields)
        else:
            cpd = db.compounds.find_one({'_id': cpd_id}, excluded_fields)
        # New MINEs won't have this precomputed
        if cpd and 'Reactant_in' not in cpd and 'Product_of' not in cpd:
            rxns_as_sub = db.reactions.find({'Reactants.c_id': cpd['_id']})
            cpd['Reactant_in'] = [x['_id'] for x in rxns_as_sub]
            rxns_as_prod = db.reactions.find({'Products.c_id': cpd['_id']})
            cpd['Product_of'] = [x['_id'] for x in rxns_as_prod]
        compounds.append(cpd)
    return compounds


def get_rxns(db, id_list):
    """Returns reactions with associated IDs from a Mongo database.

    :param db: DB to search
    :type db: A Mongo Database
    :param id_list: IDs to get compounds for.
    :type id_list: list
    :return: Reaction JSON documents
    :rtype: list
    """
    reactions = []
    for rxn_id in id_list:
        reactions.append(db.reactions.find_one({'_id': rxn_id}))

    return reactions


def get_ops(db, operator_ids):
    """Returns operators from a Mongo database.

    :param db: DB to search
    :type db: A Mongo Database
    :param operator_ids: Mongo _ids or operator names (e.g. 1.1.-1.h)
    :type operator_ids: list
    :return: Operator JSON documents
    :rtype: list
    """
    if not operator_ids:
        operators = [op for op in db.operators.find()]
    else:
        operators = []
        for op_id in operator_ids:
            op = db.operators.find_one({'$or': [{'_id': op_id},
                                                {"Name": op_id}]})
            operators.append(op)

    return operators


def get_op_w_rxns(db, operator_id):
    """Returns operator with all its associated reactions.

    :param db: DB to search
    :type db: A Mongo Database
    :param operator_id: Mongo _id or operator name (e.g. 1.1.-1.h)
    :type operator_id: str
    :return: Operator JSON document (with reactions)
    :rtype: list
    """
    operator = db.operators.find_one({'$or': [{'_id': operator_id},
                                              {"Name": operator_id}]})
    if operator:
        operator['Reaction_ids'] = \
            db.reactions.find({"Operators": operator_id}).distinct("_id")
    else:
        return None

    return operator
